from databricks.feature_store.entities._feature_store_object import _FeatureStoreObject


class Tag(_FeatureStoreObject):
    def __init__(self, key: str, value: str):
        self._key = key
        self._value = value

    @property
    def key(self) -> str:
        return self._key

    @property
    def value(self) -> str:
        return self._value

    @classmethod
    def from_proto(cls, tag_proto):
        return cls(
            key=tag_proto.key,
            value=tag_proto.value,
        )



import array
import datetime
import math
import os
import shutil
import tempfile
from datetime import timedelta

import pandas as pd
import requests
import yaml
from pyspark.sql import dataframe
from pyspark.sql.functions import lit

from databricks.feature_store.online_store_spec import (
    AmazonRdsMySqlSpec,
    AzureMySqlSpec,
    AzureSqlServerSpec,
    AmazonDynamoDBSpec,
)


def reset_hive_database(spark_session, db_name):
    drop_hive_database(spark_session, db_name)
    spark_session.sql(f"CREATE DATABASE {db_name}")


def drop_hive_database(spark_session, db_name):
    spark_session.sql(f"DROP DATABASE IF EXISTS {db_name} CASCADE")


def augment_df(df, column_name, column_int_value):
    return df.withColumn(column_name, lit(column_int_value).cast("int"))


# DataFrame rows -> dict (key is PK)
def df_to_dict(the_df):
    return {r.get("id"): r for r in [r.asDict() for r in the_df.collect()]}


# Perform "upsert"-like combine operation on input dataframes in order
def combine_df_in_order(spark_session, dfs):
    head, *tail = dfs
    expected = df_to_dict(head)
    for tail_df in tail:
        expected.update(df_to_dict(tail_df))
    return spark_session.createDataFrame(expected.values(), head.schema).collect()


# TODO: Add unit tests for assert_pyspark_df_equal method
# check if two pyspark dataframes are equal
def assert_pyspark_df_equal(
    df1: dataframe,
    df2: dataframe,
    check_column_order: bool = True,
    check_row_order: bool = True,
    atol: float = 1e-5,
    time_tol: datetime.timedelta = datetime.timedelta(seconds=1),
    sort_keys: list = [],
):
    # check schema
    # pyspark_df.dtypes is a list of tuple with column name and its type
    if check_column_order:
        assert df1.dtypes == df2.dtypes
    else:
        # ignore column order, sort by column names
        assert sorted(df1.dtypes, key=lambda x: x[0]) == sorted(
            df2.dtypes, key=lambda x: x[0]
        )

    if check_row_order:
        df1_data = df1.collect()
        df2_data = df2.collect()
    else:
        df1_data = (
            df1.sort(sort_keys, ascending=False).collect()
            if sort_keys
            else sorted(df1.collect())
        )
        df2_data = (
            df2.sort(sort_keys, ascending=False).collect()
            if sort_keys
            else sorted(df2.collect())
        )

    def is_equal(c1, c2):
        if type(c1) != type(c2):
            return False
        if type(c1) == float:
            return math.isclose(c1, c2, rel_tol=atol)
        if type(c1) in [datetime.date, datetime.datetime]:
            return abs(c1 - c2) < time_tol
        if type(c1) in [list, array]:
            return all([is_equal(c1[i], c2[i]) for i in range(len(c1))])
        if type(c1) == dict:
            return is_equal(list(c1.keys()), list(c2.keys())) and is_equal(
                list(c1.values()), list(c2.values())
            )
        return c1 == c2

    for r in range(len(df1_data)):
        for c in df1.columns:
            assert is_equal(df1_data[r][c], df2_data[r][c])


# Provider names. If adding or removing a provider, you should also add it to
# PUBLISH_PROVIDERS in FeatureStoreNotebookRunner.scala
PROVIDER_AURORA = "AURORA"
PROVIDER_RDS_MYSQL = "RDS_MYSQL"
PROVIDER_AZURE_MYSQL = "AZURE_MYSQL"
PROVIDER_AZURE_MSSQL = "AZURE_MSSQL"
PROVIDER_DYNAMODB = "DYNAMODB"

# Secret keys
SECRET_KEYS_BY_PROVIDER = {
    PROVIDER_AURORA: {
        "hostname": "AURORA-hostname",
        "port": "AURORA-port",
        "user": "AURORA-user",
        "password": "AURORA-password",
    },
    PROVIDER_RDS_MYSQL: {
        "hostname": "RDS_MYSQL-hostname",
        "port": "RDS_MYSQL-port",
        "user": "RDS_MYSQL-user",
        "password": "RDS_MYSQL-password",
    },
    PROVIDER_AZURE_MYSQL: {
        "hostname": "AZURE_MYSQL-hostname",
        "port": "AZURE_MYSQL-port",
        "user": "AZURE_MYSQL-user",
        "password": "AZURE_MYSQL-password",
    },
    PROVIDER_AZURE_MSSQL: {
        "hostname": "AZURE_MSSQL-hostname",
        "port": "AZURE_MSSQL-port",
        "user": "AZURE_MSSQL-user",
        "password": "AZURE_MSSQL-password",
    },
    PROVIDER_DYNAMODB: {
        "aws_access_key_id": "DYNAMODB-access-key-id",
        "aws_secret_access_key": "DYNAMODB-secret-access-key",
    },
}

# OnlineStoreSpec classes
ONLINE_STORE_SPEC_CLASS_BY_PROVIDER = {
    PROVIDER_AURORA: AmazonRdsMySqlSpec,
    PROVIDER_RDS_MYSQL: AmazonRdsMySqlSpec,
    PROVIDER_AZURE_MYSQL: AzureMySqlSpec,
    PROVIDER_AZURE_MSSQL: AzureSqlServerSpec,
    PROVIDER_DYNAMODB: AmazonDynamoDBSpec,
}

# JDBC Drivers
JDBC_DRIVER_NAME_BY_PROVIDER = {
    PROVIDER_AURORA: "org.mariadb.jdbc.Driver",
    PROVIDER_RDS_MYSQL: "org.mariadb.jdbc.Driver",
    PROVIDER_AZURE_MYSQL: "org.mariadb.jdbc.Driver",
    PROVIDER_AZURE_MSSQL: "com.microsoft.sqlserver.jdbc.SQLServerDriver",
}


def create_online_store_spec(
    provider,
    hostname=None,
    port=None,
    user=None,
    password=None,
    database_name=None,
    table_name=None,
    driver=None,
    read_secret_prefix=None,
    write_secret_prefix=None,
    aws_region="us-west-2",
    aws_access_key_id=None,
    aws_secret_access_key=None,
    aws_session_token=None,
    ttl: timedelta = None,
):
    cls = ONLINE_STORE_SPEC_CLASS_BY_PROVIDER[provider]
    if provider == PROVIDER_DYNAMODB:
        return cls(
            region=aws_region,
            access_key_id=aws_access_key_id,
            secret_access_key=aws_secret_access_key,
            session_token=aws_session_token,
            table_name=table_name,
            read_secret_prefix=read_secret_prefix,
            write_secret_prefix=write_secret_prefix,
            ttl=ttl,
        )
    else:
        return cls(
            hostname,
            port,
            user,
            password,
            database_name,
            table_name,
            driver,
            read_secret_prefix,
            write_secret_prefix,
        )


def get_database_user(provider, hostname, user):
    if provider != PROVIDER_AZURE_MYSQL:
        return user
    # https://docs.microsoft.com/en-us/azure/mysql/howto-connection-string
    servername = hostname.split(".")[0]
    return f"{user}@{hostname}"


def get_connection_string(
    provider, database_user, password, hostname, port, database_name=None
):
    if provider != PROVIDER_AZURE_MSSQL:
        protocol = "mysql+pymysql"
    else:
        protocol = "mssql+pyodbc"
        if database_name is None:
            database_name = "master"
    conn = f"{protocol}://{database_user}:{password}@{hostname}:{port}"
    if database_name is not None:
        conn += f"/{database_name}"
    if provider == PROVIDER_AZURE_MSSQL:
        conn += "?driver=ODBC+Driver+17+for+SQL+Server&autocommit=true"
    return conn


def modify_conda_env_yaml(conda_env_yaml_path, pip_dep_prefix, pip_dep_replacement_str):
    """
    Modify conda env yaml in place to replace any pip dependencies that start with pip_dep_prefix
    with pip_dep_replacement_str.

    If there are no pip dependencies that match pip_dep_prefix, then a ValueError will be thrown.

    :param conda_env_yaml_path the fully qualified path to the yaml file that contains the conda environment spec
    :param pip_dep_prefix the prefix of the pip dependency to match against pip dependencies in the conda yml file,
        for example: "databricks-feature-lookup"
    :param pip_dep_replacement_str the new pip dependency to use as a replacement, for example: "https://bucket/file.whl"
    """
    with open(conda_env_yaml_path, "r") as target_artifact_fd:
        conda_yml_content = yaml.safe_load(target_artifact_fd.read())
        error_msg = "No matching pip dependencies found"
        if "dependencies" not in conda_yml_content:
            raise ValueError(
                f"{error_msg}, conda yaml did not contain dependencies section"
            )
        dependencies = conda_yml_content["dependencies"]
        found_pip_deps = False
        for dependency in dependencies:
            # After parsing the conda dependencies there will be two pip dependencies, one that is a
            # string that represents the pip package itself, and another one that is a dictionary that
            # contains the nested pip dependencies.  In this case we want the dictionary with the actual pip
            # dependencies.
            if isinstance(dependency, dict) and "pip" in dependency:
                found_pip_deps = True
                pip_dependencies = dependency["pip"]
                updated_pip_dependencies = []
                found_matching_pip_dep = False
                for pip_dependency in pip_dependencies:
                    if pip_dependency.startswith(pip_dep_prefix):
                        found_matching_pip_dep = True
                        updated_pip_dependencies.append(pip_dep_replacement_str)
                    else:
                        updated_pip_dependencies.append(pip_dependency)
                if not found_matching_pip_dep:
                    raise ValueError(
                        f"{error_msg}, conda yaml did not find a matching pip dependency"
                    )
                dependency["pip"] = updated_pip_dependencies
        if not found_pip_deps:
            raise ValueError(
                f"{error_msg}, conda yaml did not contain a pip dependencies section"
            )
    with open(conda_env_yaml_path, "w") as updated_target_artifact_fd:
        yaml.dump(conda_yml_content, updated_target_artifact_fd)


def modify_pip_requirements(
    pip_requirements_path, pip_dep_prefix, pip_dep_replacement_str
):
    """
    Modify the pip requirements.txt file in place to replace any dependencies that start with
    pip_dep_prefix with pip_dep_replacement_str.

    If a matching pip dependency is not found, this will raise a ValueError.

    This does not handle the following features of the requirements.txt spec:
      - Continuation lines ending in `\`.  These will be treated as separate lines.
      - Including additional files via "-r".  These will be ignored.

    Pip requirements.txt format specification:
    https://pip.pypa.io/en/stable/reference/requirements-file-format/#requirements-file-format)

    :param pip_dep_prefix the prefix of the pip dependency to match against pip dependencies in the conda yml file,
        for example: "databricks-feature-lookup"
    :param pip_dep_replacement_str the new pip dependency to use as a replacement, for example: "https://bucket/file.whl"
    """

    found_pip_dep_prefix = False

    # Copy line-by-line to a temp file, performing desired transformations to each line.
    temp_file_fd, temp_file_path = tempfile.mkstemp()
    with os.fdopen(temp_file_fd, "w") as target_file:
        with open(pip_requirements_path) as source_file:
            for line in source_file:
                if line.strip().startswith(pip_dep_prefix):
                    found_pip_dep_prefix = True
                    target_file.write(pip_dep_replacement_str)
                else:
                    target_file.write(line)

    if not found_pip_dep_prefix:
        os.remove(temp_file_path)
        raise ValueError("No matching pip dependency found")

    # Move temp file with updated contents to overwrite original file
    os.remove(pip_requirements_path)
    shutil.move(temp_file_path, pip_requirements_path)



""" Defines the HiveClient class and some utilities used by this class. """

import logging
from pyspark.sql import SparkSession
from pyspark.sql.utils import AnalysisException, CapturedException
from pyspark.sql.functions import count, lit

from databricks.feature_store.utils import utils
from databricks.feature_store.feature_table_properties import (
    PRIMARY_KEYS,
    ALL_PROPERTIES,
)

from mlflow.pyfunc import spark_udf

_logger = logging.getLogger(__name__)


class HiveClient:
    """
    Wraps the details of interacting with the Hive metastore behind an API.
    This class is used by the FeatureStoreClient.

    The hive client should be reserved for low-level hive operations and not contain any business logic
    that is unrelated to hive itself (for example, calling the catalog backend).  If you need additional
    business logic, consider using HiveClientHelper instead.

    The idea behind this layer of abstraction is that the FeatureStoreClient may later
    be changed to create tables, set properties, etc using a RESTful client talking to
    a feature catalog service.
    """

    STREAMING_ZORDER_INTERVAL = 5000

    def __init__(self):
        """
        Constructs a HiveClient that reads and writes tables from the specified
        database_name.
        """
        self._spark = SparkSession.builder.appName(
            "feature_store.hive_client"
        ).getOrCreate()

    def database_exists(self, db_name):
        return self._spark.sql(f'SHOW DATABASES LIKE "{db_name}"').count() == 1

    def table_exists(self, qualified_table_name):
        """
        Determines whether a table exists in this database.
        """
        try:
            # spark.catalog.listTables is slow, so we instead check if this SQL query
            # throws. For more information, see https://kb.databricks.com/metastore/list-tables.html
            self._spark.sql(f"SHOW TBLPROPERTIES {qualified_table_name}").collect()
        except AnalysisException:
            return False
        return True

    def has_generated_columns(self, qualified_table_name):
        """
        Determines whether an existing table contains generated columns.
        Return True if the query failed so that we reject register table.
        """
        try:
            result = self._spark.sql(
                f"SHOW CREATE TABLE {qualified_table_name}"
            ).collect()[0]
            schema_dict = result.asDict(True)
            for value in schema_dict.values():
                if "GENERATED ALWAYS AS" in value.upper():
                    return True
            return False
        except CapturedException:
            return True

    def get_delta_table_path(self, qualified_table_name):
        """
        Expects Delta table. Returns the DBFS path to the Delta table.
        Use as documented here : https://docs.delta.io/latest/delta-utility.html#detail-schema
        """
        try:
            results = (
                self._spark.sql(f"DESCRIBE DETAIL {qualified_table_name}")
                .limit(1)
                .collect()
            )
            paths = [row["location"] for row in results]
            if len(paths) == 0:
                return None
            return paths[0]
        # All pyspark exceptions inherit CapturedException as documented here :
        # https://spark.apache.org/docs/3.0.0-preview/api/python/_modules/pyspark/sql/utils.html
        # If the query failed for whatever reason we should return None here
        except CapturedException:
            return None

    @staticmethod
    def _dbfs_path_to_table_format(path):
        # Expected DBFS path for delta table "database_name.table_name" is
        # "dbfs:/path/to/files/database_name.db/table_name"
        #
        # If input path does not match this signature, return original path
        if not path.lower().startswith("dbfs:/"):
            return path
        # split by "/" and isolate the leaf file and parent db directory
        split_path = path.split("/")
        if len(split_path) < 2:
            return path
        db_path, table_name = split_path[-2:]
        if not db_path.lower().endswith(".db"):
            return path
        # Isolate database name and verify that "database_name.table_name" <--> input path
        db_name = db_path[:-3]  # everything before the last ".db" prefix
        return f"{db_name}.{table_name}"

    # TODO(zero.qu): find proper way to get the exact database name and table name of a delta table
    # Currently we extract the database name and table name of a delta table by parsing the
    # data source path. However, the path may not give sufficient information if the user uses
    # external blob storage like S3 to store the delta table. We should fix the logic here with
    # proper exception handling.
    def convert_to_table_format(self, path):
        data_source = self._dbfs_path_to_table_format(path)
        # case-sensitive comparison since DBFS supports case-sensitive files
        if self.get_delta_table_path(data_source) != path:
            return path
        return data_source

    # TODO(avesh.singh): Remove after updating FeatureStoreClient.publish_table to use Feature
    #  Catalog
    def get_primary_keys(self, qualified_table_name):
        return self.get_property(qualified_table_name, PRIMARY_KEYS)

    def get_feature_table_schema(self, feature_table_name):
        from delta.tables import DeltaTable

        feature_table = DeltaTable.forName(self._spark, feature_table_name)
        return feature_table.toDF().schema

    def empty_df(self, schema):
        return self._spark.createDataFrame([], schema)

    def create_table(
        self, qualified_table_name, schema, partition_columns=None, path=None
    ):
        """
        Creates a Delta table in the Hive metastore.
        Will throw if schema contains duplicate columns.
        """
        df = self.empty_df(schema)
        writer = (
            df.write.partitionBy(*partition_columns) if partition_columns else df.write
        )
        if path:
            writer = writer.option("path", path)
        writer.format("delta").saveAsTable(qualified_table_name)

    def delete_empty_table(self, qualified_table_name):
        """
        Drops a table from the Hive metastore only if it is empty.
        """
        if not HiveClient._df_is_empty_optimized(
            self._spark.read.table(qualified_table_name)
        ):
            raise ValueError(
                f"Attempted to delete non-empty table {qualified_table_name}."
            )
        self.delete_table(qualified_table_name)

    # This should be used VERY VERY carefully as it will delete the entire delta table
    def delete_table(self, qualified_table_name):
        """
        Drops a table from the Hive metastore.
        """
        self._spark.sql(f"DROP TABLE IF EXISTS {qualified_table_name}")

    def read_table(
        self, qualified_table_name, as_of_delta_timestamp=None, streaming=False
    ):
        """
        Reads a Delta table, optionally as of some timestamp.
        """
        if streaming and as_of_delta_timestamp:
            raise ValueError(
                "Internal error: as_of_delta_timestamp cannot be specified when"
                " streaming=True."
            )

        base_reader = (
            # By default, Structured Streaming only handles append operations. Because
            # we have a notion of primary keys, most offline feature store operations
            # are not appends. For example, FeatureStoreClient.write_table(mode=MERGE)
            # will issue a MERGE operation.
            # In order to propagate the non-append operations to the
            # readStream, we set ignoreChanges to "true".
            # For more information,
            # see https://docs.databricks.com/delta/delta-streaming.html#ignore-updates-and-deletes
            self._spark.readStream.format("delta").option("ignoreChanges", "true")
            if streaming
            else self._spark.read.format("delta")
        )

        if as_of_delta_timestamp:
            table_reader = base_reader.option("timestampAsOf", as_of_delta_timestamp)
            table_location = self.get_delta_table_path(qualified_table_name)

            # It would be more natural to do use the .table function rather than using the
            # location of the table in DBFS, for example with code like:
            # (self._spark.read.format("delta")
            #   .option('timestampAsOf', as_of_delta_timestamp).
            #   .table(qualified_table_name)
            # However this is not possible due to
            # https://databricks.atlassian.net/browse/SC-35363
            #
            # Another option here is to use the @-syntax
            # (https://docs.databricks.com/delta/delta-batch.html#-syntax), however this
            # syntax uses the weird format "yyyyMMddHHmmssSSS", and it is preferable to
            # provide a consistent timestamp string format between
            # FeatureStoreClient.read_table and the DataFrameReader timestampAsOf option.
            return table_reader.load(table_location)
        else:
            return base_reader.table(qualified_table_name)

    def df_violates_pk_constraint(self, df, keys):
        count_column_name = "databricks__internal__row_counter"
        df_aggregated = (
            df.groupBy(*keys)
            .agg(count(lit(1)).alias(count_column_name))
            .filter(f"{count_column_name} > 1")
        )
        return not HiveClient._df_is_empty_optimized(df_aggregated)

    def write_table(
        self,
        qualified_table_name,
        primary_keys,
        timestamp_keys,
        df,
        mode,
        checkpoint_location=None,
        trigger=None,
    ):
        """
        Write features.

        :return: If ``df.isStreaming``, returns a PySpark :class:`StreamingQuery <pyspark.sql.streaming.StreamingQuery>`. :obj:`None` otherwise.
        """

        if mode not in ["overwrite", "merge"]:
            raise ValueError(f"Unsupported mode '{mode}'.")

        keys = primary_keys + timestamp_keys
        zorder_keys = primary_keys[0:2] + timestamp_keys if timestamp_keys else []

        if not df.isStreaming:
            # Verify that input dataframe has unique rows per pk combination
            if self.df_violates_pk_constraint(df, keys):
                raise ValueError(
                    f"Non-unique rows detected in input dataframe for key combination {keys}."
                )

        if df.isStreaming:
            if trigger is None:
                raise ValueError("``trigger`` must be set when df.isStreaming")
            if mode == "overwrite":
                raise TypeError(
                    "API not supported for streaming DataFrame in 'overwrite' mode."
                )
            return self._merge_streaming_df_into_delta_table(
                qualified_table_name,
                primary_keys,
                timestamp_keys,
                zorder_keys,
                df,
                trigger,
                checkpoint_location,
            )
        else:
            if mode == "overwrite":
                self._write_to_delta_table(
                    qualified_table_name, timestamp_keys, df, "overwrite"
                )
            else:
                self._merge_df_into_delta_table(
                    qualified_table_name, primary_keys, timestamp_keys, df
                )
            if zorder_keys:
                self._zorder_delta_table(qualified_table_name, zorder_keys)
            return None

    def get_predict_udf(self, model_uri, result_type=None):
        kwargs = {"result_type": result_type} if result_type else {}
        return spark_udf(self._spark, model_uri, **kwargs)

    @staticmethod
    def _write_to_delta_table(delta_table_name, timestamp_keys, df, mode):
        HiveClient._validate_timestamp_key_columns(df, timestamp_keys)
        return (
            df.write.option("mergeSchema", "true")
            .format("delta")
            .mode(mode)
            .saveAsTable(delta_table_name)
        )

    @staticmethod
    def _df_is_empty_optimized(df):
        """
        Check if the Spark DataFrame is empty.
        Using limit(1) and then count to check size rather than dropping to RDD level.
        """
        return df.limit(1).count() == 0

    @staticmethod
    def _df_columns_contain_nulls(df, columns):
        """
        Check if any of the target columns in the Spark DataFrame contain null values.
        If df or columns is empty, return False.
        """
        if not columns or HiveClient._df_is_empty_optimized(df):
            return False
        null_filter = " OR ".join([f"{column} IS NULL" for column in columns])
        filtered_df = df.filter(null_filter)
        return not HiveClient._df_is_empty_optimized(filtered_df)

    @staticmethod
    def _validate_timestamp_key_columns(df, timestamp_keys):
        if HiveClient._df_columns_contain_nulls(df, timestamp_keys):
            _logger.warning("DataFrame has null values in timestamp key column.")

    def attempt_to_update_delta_table_schema(self, delta_table_name, new_schema):
        df = self.empty_df(new_schema)
        (
            df.write.option("mergeSchema", "true")
            .format("delta")
            .mode("append")
            .saveAsTable(delta_table_name)
        )

    def _generate_merge_operation(
        self, feature_table_name, primary_keys, timestamp_keys, source_df_schema
    ):
        from delta.tables import DeltaTable
        from pyspark.sql.functions import lit

        result_table_alias = "result_table"
        batch_table_alias = "updates_table"
        source_df_columns_names = source_df_schema.fieldNames()
        feature_names = self.get_feature_table_schema(feature_table_name).fieldNames()
        keys = primary_keys + timestamp_keys

        def get_update_expression(feature):
            if feature in source_df_columns_names:
                # feature column exists in source_df
                return f"{batch_table_alias}.{feature}"
            else:
                # feature column missing in source_df, use existing value
                return f"{result_table_alias}.{feature}"

        def get_insert_expression(feature):
            if feature in source_df_columns_names:
                return f"{batch_table_alias}.{feature}"
            else:
                return lit(None)

        features = set(source_df_columns_names + feature_names)
        update_expr = {feature: get_update_expression(feature) for feature in features}
        insert_expr = {feature: get_insert_expression(feature) for feature in features}

        merge_condition = " AND ".join(
            [f"{result_table_alias}.{k} = {batch_table_alias}.{k}" for k in keys]
        )

        feature_table = DeltaTable.forName(self._spark, feature_table_name)

        def merge(batch_df):
            self._validate_timestamp_key_columns(batch_df, timestamp_keys)
            return (
                feature_table.alias(result_table_alias)
                .merge(batch_df.alias(batch_table_alias), merge_condition)
                .whenNotMatchedInsert(values=insert_expr)
                .whenMatchedUpdate(set=update_expr)
                .execute()
            )

        return merge

    def _merge_df_into_delta_table(
        self,
        feature_table_name,
        primary_keys,
        timestamp_keys,
        source_df,
    ):
        merge_fn = self._generate_merge_operation(
            feature_table_name, primary_keys, timestamp_keys, source_df.schema
        )
        merge_fn(source_df)

    def _merge_streaming_df_into_delta_table(
        self,
        feature_table_name,
        primary_keys,
        timestamp_keys,
        zorder_keys,
        source_df,
        trigger,
        checkpoint_location=None,
    ):
        merge_fn = self._generate_merge_operation(
            feature_table_name, primary_keys, timestamp_keys, source_df.schema
        )

        def batch_fn(batch_df, batch_id):
            merge_fn(batch_df)
            if zorder_keys and batch_id % self.STREAMING_ZORDER_INTERVAL == 0:
                self._zorder_delta_table(feature_table_name, zorder_keys)

        options = {}
        if checkpoint_location is not None:
            options["checkpointLocation"] = checkpoint_location

        return (
            source_df.writeStream.trigger(**trigger)
            .outputMode("update")
            .foreachBatch(batch_fn)
            .options(**options)
            .start()
        )

    def _zorder_delta_table(self, feature_table_name, zorder_keys):
        self._spark.sql(
            f"OPTIMIZE {feature_table_name} ZORDER BY ({', '.join(zorder_keys)})"
        )



import json
from typing import List, Any, Optional, Dict

import boto3
import numpy as np

DYNAMODB = "dynamodb"
TABLE = "Table"
ATTRIBUTE_NAME = "AttributeName"
ATTRIBUTE_TYPE = "AttributeType"
KEY_TYPE = "KeyType"
KEY_SCHEMA = "KeySchema"
ITEM = "Item"
ITEMS = "Items"
DYNAMODB_STRING_TYPE = "S"
DYNAMODB_NUMBER_TYPE = "N"
HASH = "HASH"
RANGE = "RANGE"

PRIMARY_KEY_ATTRIBUTE_NAME_VALUE = "_feature_store_internal__primary_keys"
PRIMARY_KEY_SCHEMA = {ATTRIBUTE_NAME: PRIMARY_KEY_ATTRIBUTE_NAME_VALUE, KEY_TYPE: HASH}
TTL_KEY_ATTRIBUTE_NAME_VALUE = "_feature_store_internal__ttl_column"


def to_dynamodb_primary_key(primary_key_values: List[Any]):
    return {PRIMARY_KEY_ATTRIBUTE_NAME_VALUE: json.dumps(primary_key_values)}


def to_range_schema(timestamp_key: str):
    return {ATTRIBUTE_NAME: timestamp_key, KEY_TYPE: RANGE}


def key_schema_to_tuple(key_schema_element: Dict[str, str]):
    return (key_schema_element[ATTRIBUTE_NAME], key_schema_element[KEY_TYPE])


def key_schemas_equal(key_schema1: List[Dict], key_schema2: List[Dict]):
    # Compare two KeySchemas, ignoring order of items.
    key_schema_tuples1 = set(key_schema_to_tuple(ks) for ks in key_schema1)
    key_schema_tuples2 = set(key_schema_to_tuple(ks) for ks in key_schema2)
    return key_schema_tuples1 == key_schema_tuples2


def to_safe_select_expression(feature_names: List[str]):
    """
    Helper to create the args for safe feature selection in DynamoDB. DynamoDB projection expressions (which define the
    attributes to retrieve), cannot be used with reserved keywords or special characters (including spaces).
    e.g. "feat 1", "_feat_1", "comment" are all invalid projection expressions.

    See the Amazon documentation for more details:
    https://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_Query.html#DDB-Query-request-ExpressionAttributeNames

    To be safe, we alias all features with [#f0, #f1...]. Selecting ["feat 1", "_feat_1", "comment"] yields:
        - ProjectionExpression="#f0, #f1, #f2"
        - ExpressionAttributeNames={"#f0": "feat 1", "#f1": "_feat_1", "#f2": "comment}

    :param feature_names: List of feature names to select.
    :return: Safe DynamoDB expression args for selecting the given features.
    """
    safe_aliases = [f"#f{i}" for i in range(len(feature_names))]
    return {
        "ProjectionExpression": ", ".join(safe_aliases),
        "ExpressionAttributeNames": dict(zip(safe_aliases, feature_names)),
    }


def get_dynamodb_resource(
    access_key_id: str,
    secret_access_key: str,
    region: str,
    session_token: Optional[str] = None,
):
    return boto3.resource(
        DYNAMODB,
        aws_access_key_id=access_key_id,
        aws_secret_access_key=secret_access_key,
        region_name=region,
        aws_session_token=session_token,
    )



from databricks.feature_store.entities._feature_store_object import _FeatureStoreObject
from databricks.feature_store.entities.job import Job
from databricks.feature_store.entities.notebook import Notebook


class Consumer(_FeatureStoreObject):
    def __init__(self, features, notebook=None, job_run=None):
        self._features = features
        self._notebook = notebook
        self._job_run = job_run

    @property
    def features(self):
        return self._features

    @property
    def notebook(self):
        return self._notebook

    @property
    def job_run(self):
        return self._job_run

    @classmethod
    def from_proto(cls, consumer_proto):
        if consumer_proto.HasField("notebook"):
            return cls(
                features=list(consumer_proto.features),
                notebook=Notebook.from_proto(consumer_proto.notebook),
            )
        elif consumer_proto.HasField("job_run"):
            return cls(
                features=list(consumer_proto.features),
                job_run=Job.from_proto(consumer_proto.job_run),
            )



from databricks.feature_store.entities.feature_table import FeatureTable
from databricks.feature_store.utils import request_context
from databricks.feature_store.utils.request_context import RequestContext

import uuid

HIVE_ILLEGAL_TABLE_NAMES = [
    "inv@lidn@me",
    ".table",
    "database.",
    "database.table.col",
    "d@tabase.table",
    "`database.table`",
]

FEATURE_STORE_ILLEGAL_TABLE_NAMES = HIVE_ILLEGAL_TABLE_NAMES + ["invalidname"]


def create_test_feature_table(
    name,
    description,
    primary_keys,
    partition_cols,
    features,
    table_id=str(uuid.uuid4()),
    creation_timestamp=0,
    notebook_producers=[],
    job_producers=[],
    timestamp_keys=[],
    path_data_sources=[],
    table_data_sources=[],
    custom_data_sources=[],
):
    return FeatureTable(
        name=name,
        table_id=table_id,
        description=description,
        primary_keys=primary_keys,
        partition_columns=partition_cols,
        timestamp_keys=timestamp_keys,
        features=features,
        creation_timestamp=creation_timestamp,
        online_stores=[],
        notebook_producers=notebook_producers,
        job_producers=job_producers,
        table_data_sources=table_data_sources,
        path_data_sources=path_data_sources,
        custom_data_sources=custom_data_sources,
    )


def assert_request_context(method_calls, expected_feature_store_method_name):
    """
    Assert that every method call in the list of mock.method_call objects that is called
    with a RequestContext parameter is the expected feature_client_method_name.

    :param method_calls: a list of method calls captured by a mock.
    :param expected_feature_store_method_name: the expected feature store method name in the
    the RequestContext parameter of the captured method calls.
    """
    for method_call in method_calls:
        _, positional_args, keyword_args = method_call
        keyword_args_vals = list(keyword_args.values())
        method_params = keyword_args_vals + list(positional_args)
        for method_param in method_params:
            if method_param.__class__ == RequestContext:
                feature_store_method_name = method_param.get_header(
                    request_context.FEATURE_STORE_METHOD_NAME
                )
                assert feature_store_method_name == expected_feature_store_method_name



import logging
from typing import Set

from databricks.feature_store.catalog_client import CatalogClient
from databricks.feature_store.databricks_client import DatabricksClient
from databricks.feature_store.utils import utils
from databricks.feature_store.utils.request_context import (
    RequestContext,
)
from mlflow.utils import databricks_utils

_logger = logging.getLogger(__name__)


class CatalogClientHelper:
    """
    Helper functions that wrap calls to the catalog client with additional business logic, possibly invoking
    other clients as well (eg, DatabricksClient).
    """

    def __init__(
        self, catalog_client: CatalogClient, databricks_client: DatabricksClient
    ):
        self._catalog_client = catalog_client
        self._databricks_client = databricks_client

    def add_producers(
        self, feature_table_name, producer_action, req_context: RequestContext
    ):
        try:
            if utils.is_in_databricks_job():
                job_id = databricks_utils.get_job_id()
                if job_id:
                    job_id = int(job_id)
                    job_run_id = databricks_utils.get_job_run_id()
                    job_run_id = int(job_run_id) if job_run_id else None
                    self._catalog_client.add_job_producer(
                        feature_table_name,
                        job_id,
                        job_run_id,
                        producer_action,
                        req_context,
                    )
                else:
                    _logger.warning(
                        f"Failed to record producer in the catalog. Missing job_id ({job_id})"
                    )
            elif databricks_utils.is_in_databricks_notebook():
                notebook_path = databricks_utils.get_notebook_path()
                notebook_id = databricks_utils.get_notebook_id()
                if notebook_id:
                    notebook_id = int(notebook_id)
                    revision_id = self._databricks_client.take_notebook_snapshot(
                        notebook_path
                    )
                    revision_id = int(revision_id) if revision_id else None
                    self._catalog_client.add_notebook_producer(
                        feature_table_name,
                        notebook_id,
                        revision_id,
                        producer_action,
                        req_context,
                    )
                else:
                    _logger.warning(
                        f"Failed to record producer in the catalog. "
                        f"Missing notebook_id ({notebook_id})."
                    )
        except Exception as e:
            _logger.warning(
                f"Failed to record producer in the catalog. Exception: {e}",
                exc_info=True,
            )

    def add_consumers(self, feature_table_map, req_context: RequestContext):
        try:
            if utils.is_in_databricks_job():
                job_id = databricks_utils.get_job_id()
                job_run_id = databricks_utils.get_job_run_id()
                if job_id:
                    job_id = int(job_id)
                    job_run_id = int(job_run_id) if job_run_id else None
                    self._catalog_client.add_job_consumer(
                        feature_table_map, job_id, job_run_id, req_context
                    )
                else:
                    _logger.warning(
                        f"Failed to record consumer in the catalog. Missing job_run_id ({job_id})."
                    )
            elif databricks_utils.is_in_databricks_notebook():
                notebook_path = databricks_utils.get_notebook_path()
                notebook_id = databricks_utils.get_notebook_id()
                if notebook_id:
                    notebook_id = int(notebook_id)
                    revision_id = self._databricks_client.take_notebook_snapshot(
                        notebook_path
                    )
                    revision_id = int(revision_id) if revision_id else None
                    self._catalog_client.add_notebook_consumer(
                        feature_table_map, notebook_id, revision_id, req_context
                    )
                else:
                    _logger.warning(
                        f"Failed to record consumer in the catalog. "
                        f"Missing notebook_id ({notebook_id})."
                    )
        except Exception as e:
            _logger.warning(
                f"Failed to record consumer in the catalog. Exception: {e}",
                exc_info=True,
            )

    def add_data_sources(
        self,
        name: str,
        tables: Set[str],
        paths: Set[str],
        custom_sources: Set[str],
        req_context: RequestContext,
    ):
        try:
            self._catalog_client.add_data_sources(
                name,
                tables=list(tables),
                paths=list(paths),
                custom_sources=list(custom_sources),
                req_context=req_context,
            )
        except Exception as e:
            _logger.warning(
                f"Failed to record data sources in the catalog. Exception: {e}",
                exc_info=True,
            )



from databricks.feature_store.entities.feature_spec import FeatureSpec
from databricks.feature_store.entities.feature_table import FeatureTable
from databricks.feature_store.utils.feature_lookup_utils import join_feature_data
from pyspark.sql import DataFrame

from typing import List, Dict


class TrainingSet:
    """
    Class that defines :obj:`TrainingSet` objects.

    .. note::

       The :class:`TrainingSet` constructor should not be called directly. Instead,
       call :meth:`FeatureStoreClient.create_training_set <databricks.feature_store.client.FeatureStoreClient.create_training_set>`.
    """

    def __init__(
        self,
        feature_spec: FeatureSpec,
        df: DataFrame,
        labels: List[str],
        feature_table_metadata_map: Dict[str, FeatureTable],
        feature_table_data_map: Dict[str, DataFrame],
    ):
        """Initialize a :obj:`TrainingSet` object."""
        assert isinstance(
            labels, list
        ), f"Expected type `list` for argument `labels`. Got '{labels}' with type '{type(labels)}'."

        self._feature_spec = feature_spec
        self._df = df
        self._labels = labels
        self._feature_table_metadata_map = feature_table_metadata_map
        self._feature_table_data_map = feature_table_data_map

    @property
    def feature_spec(self) -> FeatureSpec:
        """Define a feature spec."""
        return self._feature_spec

    def load_df(self) -> DataFrame:
        """
        Load a :class:`DataFrame <pyspark.sql.DataFrame>`.

        Return a :class:`DataFrame <pyspark.sql.DataFrame>` for training.

        The returned :class:`DataFrame <pyspark.sql.DataFrame>` has columns specified
        in the ``feature_spec`` and ``labels`` parameters provided
        in :meth:`FeatureStoreClient.create_training_set <databricks.feature_store.client.FeatureStoreClient.create_training_set>`.

        :return:
           A :class:`DataFrame <pyspark.sql.DataFrame>` for training
        """
        return join_feature_data(
            feature_spec=self._feature_spec,
            df=self._df,
            extra_columns=self._labels,
            feature_table_metadata_map=self._feature_table_metadata_map,
            feature_table_data_map=self._feature_table_data_map,
        )



""" Defines the type conversion classes from pyspark to DynamoDb compatible python types. These
 utils are used for converting from pyspark types in offline feature tables to DynamoDb compatible
  python types before publishing to DynamoDb.
  DD: go/dynamodb-fs-dd
"""

import calendar
import datetime
from abc import ABC, abstractmethod
from decimal import Decimal

from pyspark.sql.types import (
    DataType,
    BooleanType,
    ShortType,
    IntegerType,
    LongType,
    FloatType,
    DoubleType,
    StringType,
    BinaryType,
    DateType,
    TimestampType,
    DecimalType,
    ArrayType,
    MapType,
)

from databricks.feature_store.utils.converter_utils import return_if_none
from databricks.feature_store.utils.dynamodb_utils import (
    DYNAMODB_STRING_TYPE,
    DYNAMODB_NUMBER_TYPE,
)

MICROSECONDS = 1000000


class DynamoDbPySparkTypeConverter(ABC):
    @staticmethod
    @abstractmethod
    def to_dynamodb(value):
        raise NotImplementedError

    @staticmethod
    def to_pyspark(value):
        # DynamoDB -> PySpark types is not required because we never need to read back values from
        # DynamoDb in Feature Store compute client, only lookup client.
        raise NotImplementedError


class DynamoDbPySparkBooleanTypeConverter(DynamoDbPySparkTypeConverter):
    @staticmethod
    @return_if_none
    def to_dynamodb(value: bool) -> int:
        if value:
            return 1
        else:
            return 0


class DynamoDbPySparkShortIntLongTypeConverter(DynamoDbPySparkTypeConverter):
    """
    DynamoDB converter for Spark's Short, Integer, and Long data types. These data types are all mapped to DynamoDB's
    number type (which can contain all three losslessly), so type specific converters are not required.
    """

    @staticmethod
    @return_if_none
    def to_dynamodb(value: int) -> int:
        return int(value)


class DynamoDbPySparkFloatDoubleTypeConverter(DynamoDbPySparkTypeConverter):
    """
    DynamoDB converter for Spark's Float and Double data types. These data types are all mapped to DynamoDB's
    number type (which can contain both losslessly), so type specific converters are not required.
    """

    @staticmethod
    @return_if_none
    def to_dynamodb(value: float) -> Decimal:
        return Decimal(str(value))


class DynamoDbPySparkStringTypeConverter(DynamoDbPySparkTypeConverter):
    @staticmethod
    @return_if_none
    def to_dynamodb(value: str) -> str:
        return str(value)


class DynamoDbPySparkTimestampTypeConverter(DynamoDbPySparkTypeConverter):
    @staticmethod
    @return_if_none
    def to_dynamodb(value: datetime.datetime) -> int:
        # datetime.timestamp function handles the timestamp conversion to UTC.
        # We store the timestamp as epoch time in microseconds in UTC in DynamoDB.
        # https://docs.python.org/3/library/datetime.html#datetime.datetime.timestamp
        return int(value.timestamp() * MICROSECONDS)


class DynamoDbPySparkDateTypeConverter(DynamoDbPySparkTypeConverter):
    @staticmethod
    @return_if_none
    def to_dynamodb(value: datetime.date) -> int:
        # calendar.timegm function handles the timestamp conversion to UTC.
        # We store the date as epoch time in seconds in UTC in DynamoDB.
        return calendar.timegm(value.timetuple())


class DynamoDbPySparkBinaryTypeConverter(DynamoDbPySparkTypeConverter):
    @staticmethod
    @return_if_none
    def to_dynamodb(value: bytearray) -> bytearray:
        """
        No converter is needed since PySpark python native representation for BinaryType is same
        as DynamoDB.
        """
        return value


class DynamoDbPySparkDecimalTypeConverter(DynamoDbPySparkTypeConverter):
    """
    DynamoDB converter for Spark's Decimal data types. No converter is needed since PySpark python
    native representation for DecimalType is same as DynamoDB. DynamoDB supports 38 digits of
    precision which is equal to Pyspark Decimal precision.
    """

    @staticmethod
    @return_if_none
    def to_dynamodb(value: Decimal) -> Decimal:
        return value


# Complex data type converters.
# Complex data type converters are stateful and stores the type converter for underlying data.


class DynamoDbPySparkArrayTypeConverter(DynamoDbPySparkTypeConverter):
    def __init__(self, element_converter) -> None:
        self._element_converter = element_converter

    @return_if_none
    def to_dynamodb(self, value: list) -> list:
        return [self._element_converter.to_dynamodb(x) for x in value]


class DynamoDbPySparkMapTypeConverter(DynamoDbPySparkTypeConverter):
    def __init__(self, key_converter, value_converter) -> None:
        self._key_converter = key_converter
        self._value_converter = value_converter

    @return_if_none
    def to_dynamodb(self, value: dict) -> dict:
        # Dynamo Map type requires string key.
        return {
            str(self._key_converter.to_dynamodb(k)): self._value_converter.to_dynamodb(
                v
            )
            for k, v in value.items()
        }


# Helpers to get the appropriate data type converter

BASIC_DATA_TYPE_CONVERTERS = {
    BooleanType: DynamoDbPySparkBooleanTypeConverter,
    TimestampType: DynamoDbPySparkTimestampTypeConverter,
    DateType: DynamoDbPySparkDateTypeConverter,
    StringType: DynamoDbPySparkStringTypeConverter,
    FloatType: DynamoDbPySparkFloatDoubleTypeConverter,
    DoubleType: DynamoDbPySparkFloatDoubleTypeConverter,
    ShortType: DynamoDbPySparkShortIntLongTypeConverter,
    IntegerType: DynamoDbPySparkShortIntLongTypeConverter,
    LongType: DynamoDbPySparkShortIntLongTypeConverter,
    BinaryType: DynamoDbPySparkBinaryTypeConverter,
    DecimalType: DynamoDbPySparkDecimalTypeConverter,
}

COMPLEX_DATA_TYPE_CONVERTERS = {
    ArrayType: DynamoDbPySparkArrayTypeConverter,
    MapType: DynamoDbPySparkMapTypeConverter,
}


def get_type_converter(data_type: DataType) -> DynamoDbPySparkTypeConverter:
    """Helper utility that maps pyspark types to DynamoDB converter classes."""
    high_level_data_type = type(data_type)
    if high_level_data_type in BASIC_DATA_TYPE_CONVERTERS:
        return BASIC_DATA_TYPE_CONVERTERS[high_level_data_type]
    elif high_level_data_type in COMPLEX_DATA_TYPE_CONVERTERS:
        if high_level_data_type == ArrayType:
            return COMPLEX_DATA_TYPE_CONVERTERS[high_level_data_type](
                get_type_converter(data_type.elementType)
            )
        elif high_level_data_type == MapType:
            return COMPLEX_DATA_TYPE_CONVERTERS[high_level_data_type](
                get_type_converter(data_type.keyType),
                get_type_converter(data_type.valueType),
            )
    raise ValueError(f"Unsupported data type for DynamoDB publish: {data_type}")


def get_timestamp_key_type(data_type):
    if isinstance(data_type, StringType):
        return DYNAMODB_STRING_TYPE
    elif (
        isinstance(data_type, IntegerType)
        or isinstance(data_type, FloatType)
        or isinstance(data_type, DoubleType)
        or isinstance(data_type, LongType)
        or isinstance(data_type, TimestampType)
        or isinstance(data_type, DateType)
    ):
        return DYNAMODB_NUMBER_TYPE

    # Unsupported types for timestamp_keys: boolean, Struct, Map
    raise TypeError("Unsupported data type.")



import databricks.feature_store.utils.validation_utils
from databricks.feature_store.catalog_client import CatalogClient
from databricks.feature_store.online_store_spec import OnlineStoreSpec
from mlflow.protos.databricks_pb2 import ErrorCode, RESOURCE_DOES_NOT_EXIST

from typing import List, Union, Dict, Any, Optional

from databricks.feature_store.constants import (
    OVERWRITE,
    MERGE,
    _DEFAULT_PUBLISH_STREAM_TRIGGER,
)
from databricks.feature_store.online_store_publish_client_factory import (
    get_online_store_publish_client,
)
from databricks.feature_store.utils import request_context
from databricks.feature_store.utils.publish_utils import (
    update_online_store_spec_sticky_ttl,
)
from databricks.feature_store.utils.request_context import (
    RequestContext,
    PUBLISH_AUTH_TYPE,
)
from databricks.feature_store.utils import utils
from pyspark.sql.streaming import StreamingQuery
import logging
from databricks.feature_store.online_store_publish_client import is_ttl_spec
from databricks.feature_store.entities.online_store_metadata import (
    OnlineStoreMetadata,
    augment_online_store,
)
from databricks.feature_store.utils.rest_utils import get_error_code
from databricks.feature_store.hive_client import HiveClient
from databricks.feature_store.utils import schema_utils
from databricks.feature_store._hive_client_helper import HiveClientHelper

_logger = logging.getLogger(__name__)


class PublishClient:

    _PUBLISH_MODES = [OVERWRITE, MERGE]

    def __init__(
        self,
        catalog_client: CatalogClient,
        hive_client: HiveClient,
        hive_client_helper: HiveClientHelper,
    ):
        self._catalog_client = catalog_client
        self._hive_client = hive_client
        self._hive_client_helper = hive_client_helper

    def _publish_table(
        self,
        name: str,
        online_store: OnlineStoreSpec,
        *,
        filter_condition: Optional[str] = None,
        mode: str = MERGE,
        streaming: bool = False,
        checkpoint_location: Optional[str] = None,
        trigger: Dict[str, Any] = _DEFAULT_PUBLISH_STREAM_TRIGGER,
        features: Union[str, List[str], None] = None,
    ) -> Optional[StreamingQuery]:

        req_context = RequestContext(request_context.PUBLISH_TABLE)

        # Check: valid mode used
        mode_string = mode.strip().lower()
        if mode_string not in self._PUBLISH_MODES:
            supported_modes_list = ", ".join([f"'{m}'" for m in self._PUBLISH_MODES])
            raise ValueError(
                f"Unsupported mode '{mode}'. Use one of ({supported_modes_list})"
            )

        # Check: arguments and combination of arguments
        if streaming and mode == OVERWRITE:
            raise ValueError(
                f'Streaming publish_table does not support mode="{OVERWRITE}"'
            )

        # Empty input_features means the default behavior of publishing the entire table.
        input_features = utils.as_list(features, [])
        if input_features and mode == OVERWRITE:
            raise ValueError(
                f'publish_table for selected columns does not support mode="{OVERWRITE}"'
            )

        checkpoint_location = databricks.feature_store.utils.validation_utils.standardize_checkpoint_location(
            checkpoint_location
        )
        if checkpoint_location is not None and not streaming:
            _logger.warning("Ignoring checkpoint_location, since df is not streaming.")
            checkpoint_location = None

        # Check: feature table exists as a delta table
        self._hive_client_helper.check_feature_table_exists_in_hive(name)

        # Check: feature tables exists in the catalog
        if not self._catalog_client.feature_table_exists(name, req_context):
            raise ValueError(f"Feature table '{name}' does not exists in the catalog.")

        catalog_feature_table = self._catalog_client.get_feature_table(
            name, req_context
        )

        # Check: schema between delta table and in catalog
        hive_feature_table_df = self._hive_client.read_table(name, None)
        catalog_features = self._catalog_client.get_features(name, req_context) or []

        # ToDo(mparkhe): DRY this code with schema comparison in write_tables
        # Check: Primary keys
        if not schema_utils.catalog_matches_delta_schema(
            catalog_features,
            hive_feature_table_df.schema,
            column_filter=catalog_feature_table.primary_keys or [],
        ):
            primary_keys = [
                pk
                for pk in catalog_features
                if pk.name in catalog_feature_table.primary_keys
            ]
            raise ValueError(
                f"Mismatched primary keys. Primary keys in the catalog, '{primary_keys}' do not "
                f"match keys in the data table schema {hive_feature_table_df.schema}."
            )

        # Check: Partition columns
        if (
            catalog_feature_table.partition_columns
            and not schema_utils.catalog_matches_delta_schema(
                catalog_features,
                hive_feature_table_df.schema,
                column_filter=catalog_feature_table.partition_columns or [],
            )
        ):
            partition_columns = [
                pk
                for pk in catalog_features
                if pk.name in catalog_feature_table.partition_columns
            ]
            raise ValueError(
                f"Mismatched partition columns. Partition columns  in the catalog, "
                f"'{partition_columns}' do not match keys in the data table "
                f"schema {hive_feature_table_df.schema}."
            )

        catalog_feature_names = {feature.name for feature in catalog_features}

        features_to_publish = None
        if input_features:
            input_features_names_lower_case = [
                feature_name.lower() for feature_name in input_features
            ]
            catalog_feature_names_lower_case = [
                feature_name.lower() for feature_name in catalog_feature_names
            ]
            input_features_not_in_catalog = [
                feature_name
                for feature_name in input_features_names_lower_case
                if feature_name not in catalog_feature_names_lower_case
            ]

            if len(input_features_not_in_catalog) > 0:
                raise ValueError(
                    f"Unknown feature column names. Input features, "
                    f"'{input_features_not_in_catalog}' not found in Feature Store."
                )
            input_features_correct_case = [
                feature_name
                for feature_name in catalog_feature_names
                if feature_name.lower() in input_features_names_lower_case
            ]
            # Add required primary key and timestamp key columns to user input features if they are not already included.
            features_to_publish = list(
                set(input_features_correct_case).union(
                    set(catalog_feature_table.keys),
                    set(catalog_feature_table.timestamp_keys),
                )
            )

        # Check: features listed in the catalog for publish are also available in the delta table
        catalog_feature_names_for_publish = (
            features_to_publish if features_to_publish else catalog_feature_names
        )
        features_for_publish_not_in_hive = [
            feature_name
            for feature_name in catalog_feature_names_for_publish
            if feature_name not in hive_feature_table_df.columns
        ]
        if len(features_for_publish_not_in_hive) > 0:
            raise RuntimeError(
                f"Schema mismatch. The following features being published were found in Feature Store, "
                f"but not in the data table: {sorted(features_for_publish_not_in_hive)}."
            )

        # check: features listed in the catalog not in the delta table and warn.
        # It warns when the selected columns we publish pass the previous check but there is a
        # mismatch for the unpublished columns.
        features_not_in_hive = [
            feature_name
            for feature_name in catalog_feature_names
            if feature_name not in hive_feature_table_df.columns
        ]
        if len(features_not_in_hive) > 0:
            _logger.warning(
                f"Schema mismatch. {features_not_in_hive} found in the feature catalog, "
                f"but were not found in the data table."
            )

        # Check: features not found catalog and warn.
        features_not_in_catalog = [
            feature_name
            for feature_name in hive_feature_table_df.columns
            if feature_name not in catalog_feature_names
        ]
        if len(features_not_in_catalog) > 0:
            if features_to_publish:
                _logger.warning(
                    f"Schema mismatch. {features_not_in_catalog} not found in the catalog. "
                )
            else:
                _logger.warning(
                    f"Schema mismatch. {features_not_in_catalog} not found in the catalog. "
                    f"Extra features in data table will not be published to the online store."
                )

        # Check: Time series publish attributes can only be used with time series feature tables.
        # This only needs to be checked prior to the first publish, as timestamp keys and ttl are immutable.
        # TTL is allowed to be None for both regular and time series (snapshot publish) feature tables.
        if is_ttl_spec(online_store) and online_store.ttl is not None:
            if len(catalog_feature_table.timestamp_keys) == 0:
                raise ValueError(
                    "Time to live can only be used with time series feature tables."
                )

        # Prepare DataFrame for publish
        publish_df = self._hive_client.read_table(
            name, as_of_delta_timestamp=None, streaming=streaming
        )

        # Filter rows
        if filter_condition:
            # force filtering before proceeding with other operations.
            # e.g. computing latest snapshot
            publish_df = publish_df.filter(filter_condition).cache()

        # Only publish features that exist in the catalog.
        if len(features_not_in_catalog) > 0:
            selected_columns = [
                c for c in publish_df.columns if c not in features_not_in_catalog
            ]
            publish_df = publish_df.select(*selected_columns)

        # Only publish features that are selected by user.
        if features_to_publish:
            user_selected_columns = [
                c for c in publish_df.columns if c in features_to_publish
            ]
            _logger.info(
                f"The following features will be published to the online store: [{user_selected_columns}]."
            )
            publish_df = publish_df.select(*user_selected_columns)

        # Augment the online store spec (e.g. resolve and store implicit values as part of the spec)
        online_store = augment_online_store(online_store, name)

        # TODO (ML-22021): validate that there is only one timestamp key prior to publishing
        os_publish_client = get_online_store_publish_client(online_store)
        online_table = os_publish_client.get_or_create_online_table(
            publish_df,
            catalog_feature_table.primary_keys,
            catalog_feature_table.timestamp_keys,
        )

        # Update catalog metadata with publish log
        try:
            # Create the online store metadata object for use in publish
            online_store_metadata = OnlineStoreMetadata(
                online_store, online_table.cloud_provider_unique_id
            )

            # Call GetOnlineStore and retrieve the OnlineStoreDetailed from the Feature Catalog.
            # Then, update the online store spec with any relevant sticky values returned (e.g. ttl).
            # If no online store is found, nothing needs to be done since this is a new store.
            try:
                online_store_detailed = self._catalog_client.get_online_store(
                    name, online_store_metadata, req_context
                )
                online_store = update_online_store_spec_sticky_ttl(
                    online_store, online_store_detailed
                )
                # Update the online store metadata object to reflect the updates for the online store spec.
                online_store_metadata = OnlineStoreMetadata(
                    online_store, online_table.cloud_provider_unique_id
                )
            except Exception as e:
                # All exceptions not caused by a non-existent online store should be re-raised.
                if get_error_code(e) != ErrorCode.Name(RESOURCE_DOES_NOT_EXIST):
                    raise e

            publish_req_context = RequestContext(
                request_context.PUBLISH_TABLE,
                # Track the publish auth type.
                {PUBLISH_AUTH_TYPE: online_store.auth_type()},
            )
            # Resolve sticky values prior to calling PublishFeatureTable, as the `None` value is ambiguous.
            self._catalog_client.publish_feature_table(
                name,
                online_store_metadata,
                features_to_publish,
                publish_req_context,
            )
        except Exception as e:
            if online_table.is_new_empty_table:
                additional_msg = (
                    f"You may need to manually clean up the empty table '{online_table.name}' "
                    f"at the specified online store location. "
                )
            else:
                additional_msg = ""
            _logger.error(
                f"Failed to record online store in the catalog. {additional_msg}Exception: {e}",
                exc_info=True,
            )
            raise e

        # The lookback_window defaults to the online store TTL
        lookback_window = online_store.ttl if is_ttl_spec(online_store) else None

        # Create a new publish client, since the online store spec may have been updated.
        os_publish_client = get_online_store_publish_client(online_store)
        return os_publish_client.publish(
            publish_df,
            catalog_feature_table.primary_keys,
            catalog_feature_table.timestamp_keys,
            streaming,
            mode_string,
            trigger,
            checkpoint_location,
            lookback_window=lookback_window,
        )



import time
from datetime import timedelta

from databricks.feature_store.online_store_spec import (
    OnlineStoreSpec,
    AmazonRdsMySqlSpec,
    AzureSqlServerSpec,
    AzureMySqlSpec,
)
from databricks.feature_store.publish_engine import (
    PublishMySqlEngine,
    PublishSqlEngine,
    PublishSqlServerEngine,
)
from databricks.feature_store.online_store_publish_client import (
    OnlineTable,
    OnlineStorePublishClient,
    is_rdbms_spec,
)
from databricks.feature_store.utils.publish_utils import get_latest_snapshot
from pyspark.sql import SparkSession
from py4j.protocol import Py4JJavaError
import random
import string
from databricks.feature_store.utils.spark_utils import serialize_complex_data_types
from typing import List, Optional
from databricks.feature_store.entities.feature import Feature


def temp_table_name():
    rand_suffix = "".join(random.choices(string.ascii_lowercase + string.digits, k=8))
    return f"databricks_tmp_{rand_suffix}"


def generate_sql_engine(online_store_spec, spark_session):
    if isinstance(online_store_spec, AzureSqlServerSpec):
        return PublishSqlServerEngine(online_store_spec, spark_session)
    return PublishMySqlEngine(online_store_spec, spark_session)


class OnlineStorePublishRdbmsClient(OnlineStorePublishClient):
    def __init__(self, online_store):
        if not is_rdbms_spec(online_store):
            raise ValueError(f"Unexpected online store type {type(online_store)}")
        self.database = online_store.database_name
        self.table_name = online_store.table_name
        self.online_store = online_store
        self.spark_session = SparkSession.builder.appName(
            "feature_store.rdbms_client"
        ).getOrCreate()

        self.sql_engine = generate_sql_engine(self.online_store, self.spark_session)

    def _online_table_exists(self, table_name):
        # TODO: Check if this works for a user without read access to table_name.
        result_set = self.sql_engine.get_online_tables(table_names=[table_name])
        return result_set.next()  # Returns false if there are no rows in the result set

    def _get_column_types(self, table_name):
        result_set = self.sql_engine.get_column_types(table_name)

        col_to_type = {}

        while result_set.next():
            data_type = result_set.getString(PublishSqlEngine.DATA_TYPE)
            # JDBC writes string columns in Spark dataframes to MSSQL as NVARCHAR(max) columns
            # because TEXT is deprecated in MSSQL. Therefore, we append the string size when
            # returning the NVARCHAR column type.
            if data_type == "nvarchar":
                data_type += "(max)"
            # JDBC writes binary columns in Spark dataframes to MSSQL as VARBINARY columns
            # because the maximum limit for BLOB is 50 in MSSQL. Therefore, we append the max
            # size when returning the NVARCHAR column type.
            elif data_type == "varbinary":
                data_type += "(max)"
            # Decimal columns in MySQL and MSSQL require precision and scale information, which needs
            # to be fetched separately from the numeric_precision and numeric_scale columns in the
            # database's information schema table.
            elif data_type == "decimal":
                precision = result_set.getString(PublishSqlEngine.NUMERIC_PRECISION)
                scale = result_set.getString(PublishSqlEngine.NUMERIC_SCALE)
                data_type = f"decimal({precision},{scale})"
            col_to_type[result_set.getString(PublishSqlEngine.COLUMN_NAME)] = data_type

        return col_to_type

    def _extend_schema(self, base_sql_table, new_sql_table):
        base_columns_to_types = self._get_column_types(base_sql_table)
        new_columns_to_types = self._get_column_types(new_sql_table)

        add_columns = []

        for (col, type) in new_columns_to_types.items():
            if col in base_columns_to_types and type != base_columns_to_types[col]:
                raise ValueError(
                    "Existing data in the online store is incompatible with the data "
                    "being published. To resolve this error, drop the table "
                    f'"{base_sql_table}" from the online store.'
                )
            if col not in base_columns_to_types:
                add_columns.append((col, type))

        self.sql_engine.add_columns(base_sql_table, add_columns)

    def _overwrite_table(self, original_sql_table, tmp_sql_table):
        """
        Overwrite the original table with a temporary one through a drop and a rename.
        Manual error handling is required for MySQL as DROP and RENAME commands cannot be rolled back.
        """
        try:
            with self.sql_engine.in_transaction():
                self.sql_engine.drop_table(original_sql_table)
                self.sql_engine.rename_table(tmp_sql_table, original_sql_table)
        except Exception as e:
            # Both the temporary and original table exist in the case of a SQL Server error
            if isinstance(self.sql_engine, PublishSqlServerEngine):
                self.sql_engine.drop_table(tmp_sql_table)
                raise ValueError(
                    f"Could not overwrite table {original_sql_table} due to error: {e}. "
                    f"No data has been modified."
                )
            # The temporary table is guaranteed to exist in the case of a MySQL error
            elif isinstance(self.sql_engine, PublishMySqlEngine):
                raise ValueError(
                    f"Could not rename table {tmp_sql_table} to {original_sql_table} due to error: {e}. "
                    f"Please execute this manually."
                )

    def _publish_overwrite(
        self,
        df,
        primary_keys: List[str],
        timestamp_keys,
        lookback_window: Optional[timedelta],
    ):
        # Lookback window and publishing a range of data is not supported for SQL stores.
        if lookback_window:
            raise ValueError(
                "Publishing with a lookback window is not supported for SQL online stores."
            )
        if timestamp_keys:
            df = get_latest_snapshot(df, primary_keys, timestamp_keys)
        if not self._online_table_exists(self.table_name):
            self._create_table(
                self.table_name,
                df,
                self.sql_engine.jdbc_url,
                self.sql_engine.jdbc_properties,
                primary_keys,
            )
        else:
            tmp_table = self._write_to_tmp_table(
                df,
                self.sql_engine.jdbc_url,
                self.sql_engine.jdbc_properties,
                primary_keys,
            )
            self._overwrite_table(self.table_name, tmp_table)

    def _extend_schema_and_merge_into(
        self, base_sql_table, source_sql_table, columns, primary_keys
    ):
        """
        Extend the base table's schema to match the source table's, then merges the source table into the base.
        On failure, the operation is retried to mitigate potential data inconsistencies.
        For example, MySQL schema extensions cannot be rolled back so a merge failure will lead to NULL data.
        """
        max_tries = 2
        for tries in range(1, max_tries + 1):
            try:
                with self.sql_engine.in_transaction():
                    self._extend_schema(base_sql_table, source_sql_table)
                    self.sql_engine.merge_table_into(
                        base_sql_table, source_sql_table, columns, primary_keys
                    )
                break
            except:
                if tries == max_tries:
                    raise
                time.sleep(2)

    def _publish_merge(
        self,
        df,
        primary_keys: List[str],
        timestamp_keys,
        lookback_window: Optional[timedelta],
    ):
        """
        DataFrameWriter.jdbc does not support the write semantics we want (described in
        `publish` docstring). Thus this function initially writes all data from df into
        a temporary SQL table. Then SQL code is run to insert this data into the
        online store.
        """
        # Lookback window and publishing a range of data is not supported for SQL stores.
        if lookback_window:
            raise ValueError(
                "Publishing with a lookback window is not supported for SQL online stores."
            )
        if timestamp_keys:
            df = get_latest_snapshot(df, primary_keys, timestamp_keys)
        if not self._online_table_exists(self.table_name):
            self._create_table(
                self.table_name,
                df,
                self.sql_engine.jdbc_url,
                self.sql_engine.jdbc_properties,
                primary_keys,
            )
        else:
            tmp_table = self._write_to_tmp_table(
                df,
                self.sql_engine.jdbc_url,
                self.sql_engine.jdbc_properties,
                primary_keys,
            )

            try:
                self._extend_schema_and_merge_into(
                    self.table_name, tmp_table, df.columns, primary_keys
                )
            finally:
                self.sql_engine.drop_table(tmp_table)

    def _create_table(
        self, name, df, jdbc_url, connection_properties, primary_keys: str
    ):
        """
        Creates and writes df to SQL table `name` defined by connection_properties, sets
        the primary key(s).
        """
        self.sql_engine.create_empty_table(
            name, df.schema, jdbc_url, connection_properties, primary_keys
        )
        try:
            df = serialize_complex_data_types(df)
            df.write.jdbc(
                url=jdbc_url,
                table=name,
                mode="append",
                properties=connection_properties,
            )
        except Py4JJavaError as e:
            error_str = str(e)
            # This write may fail if the string primary key maximum length is exceeded.
            if "java.sql.BatchUpdateException" in error_str and (
                PublishMySqlEngine.PRIMARY_KEY_STRING_ERROR in error_str
                or PublishSqlServerEngine.PRIMARY_KEY_STRING_ERROR in error_str
            ):
                raise ValueError(
                    "Unable to publish features because a primary key string is too long. "
                    "Check the online store spec documentation for maximum primary key string lengths."
                )
            else:
                raise

    def _write_to_tmp_table(
        self, df, jdbc_url, connection_properties, primary_keys: str
    ):
        """
        Writes df to a temporary SQL table defined by connection_properties, sets
        the primary key(s).
        """
        name = temp_table_name()
        self._create_table(name, df, jdbc_url, connection_properties, primary_keys)
        return name

    def get_or_create_online_table(self, df, primary_keys, timestamp_keys):
        # RDBMS online stores support both "overwrite" and "merge" mode for publish and
        # creating a table instance is tightly ingrained as a part of the publish logic.
        # For now this method is no-op. It does not create a new online table if one does not exist
        # and returns None for cloud provider ID.
        # Logic for creation for empty table is handled in the specific publish functions.
        # ToDo(ML-20621): Identify the need to support "overwrite" mode in future and design
        #                 how online table can be created in advance to save the cloud provider
        #                 unique ID in backend catalog.
        return OnlineTable(
            name=self.table_name, cloud_provider_unique_id=None, new_table=False
        )

    def close(self):
        """
        Closes the sql engine connection.
        :return:
        """
        self.sql_engine.close()



SOURCE_DATA_COLUMN_INFO = "source_data_column_info"
FEATURE_COLUMN_INFO = "feature_column_info"
FEATURE_STORE = "feature_store"
INPUT_COLUMNS = "input_columns"
NAME = "name"
OUTPUT_NAME = "output_name"
SOURCE = "source"
TRAINING_DATA = "training_data"
INPUT_TABLES = "input_tables"
TABLE_NAME = "table_name"
TABLE_ID = "table_id"
SERIALIZATION_VERSION = "serialization_version"



""" Spark utility function. """

from pyspark.sql.types import ShortType
from pyspark.sql.functions import to_json
from databricks.feature_store.constants import COMPLEX_DATA_TYPES
from databricks.feature_store.entities.data_type import DataType


def _get_columns_of_spark_type(schema, spark_type):
    return [
        field.name for field in schema.fields if isinstance(field.dataType, spark_type)
    ]


def get_columns_of_type_short(schema):
    return _get_columns_of_spark_type(schema, ShortType)


def serialize_complex_data_types(df):
    """
    Serialize all columns in df that are of complex datatype using Spark's to_json() udf.
    """
    schema = df.schema
    select_column_list = [
        to_json(column).alias(column)
        if DataType.from_spark_type(schema[column].dataType) in COMPLEX_DATA_TYPES
        else column
        for column in df.columns
    ]
    return df.select(select_column_list)



from databricks.feature_store.entities._feature_store_object import _FeatureStoreObject


class Job(_FeatureStoreObject):
    def __init__(self, job_id, run_id, job_workspace_id, feature_table_workspace_id):
        self._job_id = job_id
        self._run_id = run_id
        self._job_workspace_id = job_workspace_id
        self._feature_table_workspace_id = feature_table_workspace_id

    @property
    def job_id(self):
        return self._job_id

    @property
    def run_id(self):
        return self._run_id

    @property
    def job_workspace_id(self):
        return self._job_workspace_id

    @property
    def feature_table_workspace_id(self):
        return self._feature_table_workspace_id

    @classmethod
    def from_proto(cls, job_proto):
        return cls(
            job_id=job_proto.job_id,
            run_id=job_proto.run_id,
            job_workspace_id=job_proto.job_workspace_id,
            feature_table_workspace_id=job_proto.feature_table_workspace_id,
        )



""" SQL utility functions """
import re

TABLE_NAME_REGEX = r"^([\w_]+\.[\w_]+)$"


def is_qualified_table_name(qualified_name):
    """
    Returns True if qualified_name is of the form <database_name>.<table_name>
    and only contain alphabet characters, numbers and _.

    This character set is required by Hive and Spark SQL:
    https://docs.databricks.com/spark/latest/spark-sql/language-manual/sql-ref-identifiers.html
    http://hive.apache.org/javadocs/r3.0.0/api/org/apache/hadoop/hive/metastore/utils/MetaStoreUtils.html#validateName-java.lang.String-org.apache.hadoop.conf.Configuration-
    """
    return re.match(TABLE_NAME_REGEX, qualified_name)



import time
from datetime import timedelta
from typing import List, Optional

from pyspark.sql import SparkSession
from pyspark.sql.functions import unix_timestamp

from databricks.feature_store.constants import OVERWRITE
from databricks.feature_store.online_store_spec import (
    OnlineStoreSpec,
    AmazonDynamoDBSpec,
    AzureCosmosDBSpec,
)
from databricks.feature_store.online_store_publish_client import (
    OnlineTable,
    OnlineStorePublishClient,
    is_nosql_spec,
)
from databricks.feature_store.publish_engine import (
    PublishDynamoDBEngine,
    PublishCosmosDBEngine,
)
from databricks.feature_store.utils.publish_utils import get_latest_snapshot


def generate_nosql_engine(online_store_spec, spark_session):
    if isinstance(online_store_spec, AmazonDynamoDBSpec):
        return PublishDynamoDBEngine(online_store_spec, spark_session)
    elif isinstance(online_store_spec, AzureCosmosDBSpec):
        return PublishCosmosDBEngine(online_store_spec, spark_session)


class OnlineStorePublishNoSqlClient(OnlineStorePublishClient):
    def __init__(self, online_store: OnlineStoreSpec):
        if not is_nosql_spec(online_store):
            raise ValueError(f"Unexpected online store type {type(online_store)}")
        spark_session = SparkSession.builder.appName(
            "feature_store.nosql_client"
        ).getOrCreate()
        self.nosql_engine = generate_nosql_engine(online_store, spark_session)

    def get_or_create_online_table(self, df, primary_keys, timestamp_keys):
        cloud_provider_unique_id = self.nosql_engine.get_cloud_provider_unique_id()
        if not cloud_provider_unique_id:
            cloud_provider_unique_id = self.nosql_engine.create_empty_table(
                df.schema, primary_keys, timestamp_keys
            )
            new_table = True
        else:
            new_table = False
        return OnlineTable(
            self.nosql_engine.table_name, cloud_provider_unique_id, new_table
        )

    def _publish_merge_range_query(
        self,
        df,
        primary_keys: List[str],
        timestamp_keys: List[str],
        lookback_window: timedelta,
    ):
        """
        Helper to publish a window of a time series dataframe to the Online Store (used with QueryMode: RANGE_QUERY).
        The timestamp key should be defined as part of the logical primary key
        (e.g. the range key in DynamoDB, id property for Cosmos DB).
        """
        # Filter the df to all data since the start of the lookback window,
        # adding a day to the window to be safe as the database will automatically remove expired records.
        utc_timestamp = time.time()
        lookback_with_buffer = (
            lookback_window.total_seconds() + timedelta(days=1).total_seconds()
        )
        df = df.filter(
            unix_timestamp(df[timestamp_keys[0]]) > utc_timestamp - lookback_with_buffer
        )

        # Generate TTL column for the online store if required
        df = self.nosql_engine.generate_df_with_ttl_if_required(df, timestamp_keys)
        self.nosql_engine.write_table(df, df.schema, primary_keys, timestamp_keys)

    def _publish_merge_primary_keys_lookup(
        self, df, primary_keys: List[str], timestamp_keys: List[str]
    ):
        """
        Helper to publish a dataframe to the Online Store. For time series dataframes, this will publish the latest snapshot.
        The timestamp key should not be a part of the logical primary key:
        (e.g. the range key in DynamoDB, id property for Cosmos DB).
        """
        # TODO (ML-21776): Add test coverage for NoSQL snapshot tables
        if timestamp_keys:
            df = get_latest_snapshot(df, primary_keys, timestamp_keys)
        # For time series tables, `get_latest_snapshot` will retrieve the latest value for each primary key.
        # Because of this deduplication, timestamp keys are not required and instead treated as a regular feature.
        self.nosql_engine.write_table(df, df.schema, primary_keys, [])

    def _publish_overwrite(
        self,
        df,
        primary_keys: List[str],
        timestamp_keys: List[str],
        lookback_window: Optional[timedelta],
    ):
        raise NotImplementedError(f'Publish mode "{OVERWRITE}" is not supported.')

    def _publish_merge(
        self,
        df,
        primary_keys: List[str],
        timestamp_keys: List[str],
        lookback_window: Optional[timedelta],
    ):
        """
        Writes all the data from df into the Online Store. Existing data will be merged.
        """
        # If the lookback window is defined, we publish a range of data for use in QueryMode RANGE_QUERY.
        # Otherwise, we publish the data as-is for non-time series feature tables and the latest snapshot for time series
        # feature tables, both of which will be used in QueryMode PRIMARY_KEY_LOOKUP.
        if lookback_window is not None:
            self._publish_merge_range_query(
                df, primary_keys, timestamp_keys, lookback_window
            )
        else:
            self._publish_merge_primary_keys_lookup(df, primary_keys, timestamp_keys)

    def close(self):
        """
        Closes the nosql engine connection.
        :return:
        """
        self.nosql_engine.close()



import logging
import copy
from databricks.feature_store.entities._feature_store_object import _FeatureStoreObject
from typing import List, Union, Optional, Dict
from databricks.feature_store.utils import utils

_logger = logging.getLogger(__name__)


class FeatureLookup(_FeatureStoreObject):
    """Value class used to specify a feature to use in a :class:`TrainingSet <databricks.feature_store.training_set.TrainingSet>`.

    :param table_name: Feature table name.
    :param lookup_key: Key to use when joining this feature table with the :class:`DataFrame <pyspark.sql.DataFrame>` passed to
       :meth:`.FeatureStoreClient.create_training_set`. The ``lookup_key`` must be the columns
       in the DataFrame passed to :meth:`.FeatureStoreClient.create_training_set`. The type of
       ``lookup_key`` columns in that DataFrame must match the type of the primary key of the
       feature table referenced in this :class:`FeatureLookup`.

    :param feature_names: A single feature name, a list of feature names, or None to lookup all features
        (excluding primary keys) in the feature table at the time that the training set is created.  If your model
        requires primary keys as features, you can declare them as independent FeatureLookups.
    :param rename_outputs: If provided, renames features in the :class:`TrainingSet <databricks.feature_store.training_set.TrainingSet>`
        returned by of :meth:`FeatureStoreClient.create_training_set <databricks.feature_store.client.FeatureStoreClient.create_training_set>`.
    :param timestamp_lookup_key: Key to use when performing point-in-time lookup on this feature table
        with the :class:`DataFrame <pyspark.sql.DataFrame>` passed to :meth:`.FeatureStoreClient.create_training_set`.
        The ``timestamp_lookup_key`` must be the columns in the DataFrame passed to :meth:`.FeatureStoreClient.create_training_set`.
        The type of ``timestamp_lookup_key`` columns in that DataFrame must match the type of the timestamp key of the
        feature table referenced in this :class:`FeatureLookup`.

          .. note::

             Experimental: This argument may change or be removed in
             a future release without warning.

    :param feature_name: Feature name.  **Deprecated** as of version 0.3.4. Use `feature_names`.
    :param output_name: If provided, rename this feature in the output of
       :meth:`FeatureStoreClient.create_training_set <databricks.feature_store.client.FeatureStoreClient.create_training_set>`.
       **Deprecated** as of version 0.3.4 . Use `rename_outputs`.
    """

    def __init__(
        self,
        table_name: str,
        lookup_key: Union[str, List[str]],
        *,
        feature_names: Union[str, List[str], None] = None,
        rename_outputs: Optional[Dict[str, str]] = None,
        timestamp_lookup_key: Union[str, List[str], None] = None,
        **kwargs,
    ):
        """Initialize a FeatureLookup object."""

        self._feature_name_deprecated = kwargs.pop("feature_name", None)
        self._output_name_deprecated = kwargs.pop("output_name", None)

        if kwargs:
            raise TypeError(
                f"FeatureLookup got unexpected keyword argument(s): {list(kwargs.keys())}"
            )

        self._table_name = table_name

        if rename_outputs is not None and not isinstance(rename_outputs, dict):
            raise ValueError(
                f"Unexpected type for rename_outputs: {type(rename_outputs)}"
            )

        self._feature_names = utils.as_list(feature_names, default=[])

        # Make sure the user didn't accidentally pass in any nested lists/dicts in feature_names
        for fn in self._feature_names:
            if not isinstance(fn, str):
                raise ValueError(
                    f"Unexpected type for element in feature_names: {type(self._feature_names)}, only strings allowed in list"
                )

        self._lookup_key = copy.copy(lookup_key)
        self._timestamp_lookup_key = copy.copy(timestamp_lookup_key)

        self._rename_outputs = {}
        if rename_outputs is not None:
            self._rename_outputs = rename_outputs.copy()

        self._inject_deprecated_feature_name()
        self._inject_deprecated_output_name()

    @property
    def table_name(self):
        """The table name to use in this FeatureLookup."""
        return self._table_name

    @property
    def lookup_key(self):
        """The lookup key(s) to use in this FeatureLookup."""
        return self._lookup_key

    @property
    def feature_name(self):
        """The feature name to use in this FeatureLookup. **Deprecated** as of version 0.3.4. Use `feature_names`."""
        return self._feature_name_deprecated

    @property
    def output_name(self):
        """The output name to use in this FeatureLookup. **Deprecated** as of version 0.3.4. Use `feature_names`."""
        if self._output_name_deprecated:
            return self._output_name_deprecated
        else:
            return self._feature_name_deprecated

    @property
    def timestamp_lookup_key(self):
        return self._timestamp_lookup_key

    def _get_feature_names(self):
        return self._feature_names

    def _get_output_name(self, feature_name):
        """Lookup the renamed output, or fallback to the feature name itself if no mapping is present"""
        return self._rename_outputs.get(feature_name, feature_name)

    def _inject_deprecated_feature_name(self):
        if self._feature_name_deprecated:
            if len(self._feature_names) > 0:
                raise ValueError(
                    "Use either feature_names or feature_name parameter, but not both."
                )
            _logger.warning(
                f'The feature_name parameter is deprecated. Use "feature_names".'
            )
            self._feature_names = [self._feature_name_deprecated]

    def _inject_deprecated_output_name(self):
        if len(self._feature_names) == 1 and self._output_name_deprecated:
            if len(self._rename_outputs) > 0:
                raise ValueError(
                    "Use either output_name or rename_outputs parameter, but not both."
                )
            _logger.warning(
                f'The output_name parameter is deprecated.  Use "rename_outputs".'
            )
            self._rename_outputs[self._feature_names[0]] = self._output_name_deprecated



from databricks.feature_store.client import FeatureStoreClient
from databricks.feature_store.constants import MERGE
from pyspark.sql import DataFrame
from pyspark.sql.streaming import StreamingQuery
from functools import wraps
from typing import List, Dict, Union, Any


def feature_table(func):  # noqa: D212,D400,D415
    """
    .. note::

       Experimental: This decorator may change or be
       removed in a future release without warning.

    The ``@feature_table`` decorator specifies that a
    function is used to generate feature data.
    Functions decorated with ``@feature_table`` must return a
    single :class:`DataFrame <pyspark.sql.DataFrame>`, which will be
    written to Feature Store. For example:

    .. code-block:: python

       from databricks.feature_store import feature_table

       @feature_table
       def compute_customer_features(data):
           '''Feature computation function that takes raw
              data and returns a DataFrame of features.'''
           return (data.groupBy('cid')
             .agg(count('*').alias('num_purchases'))
           )

    A function that is decorated with the ``@feature_table`` decorator will gain these function attributes:

    .. function:: compute_and_write(input: Dict[str, Any], feature_table_name: str, mode: str = 'merge') -> pyspark.sql.dataframe.DataFrame

       .. note::

          Experimental: This function may change or be removed in
          a future release without warning.

       Calls the decorated function using the provided ``input``, then writes
       the output :class:`DataFrame <pyspark.sql.DataFrame>` to the feature table specified by ``feature_table_name``.

       .. code-block:: python
          :caption: Example:

          compute_customer_features.compute_and_write(
              input={
                  'data': data,
              },
              feature_table_name='recommender_system.customer_features',
              mode='merge'
          )

       :param input: If ``input`` is not a dictionary,
          it is passed to the decorated function as the first positional argument.
          If ``input`` is a dictionary, the contents are unpacked and passed to
          the decorated function as keyword arguments.
       :param feature_table_name: A feature table name of the form
          ``<database_name>.<table_name>``, for example ``dev.user_features``.
          Raises exception if this feature table does not exist.
       :param mode: Two supported write modes: ``"overwrite"`` updates the whole table,
          while ``"merge"`` will upsert the rows in ``df`` into the feature table.
       :return: :class:`DataFrame <pyspark.sql.DataFrame>` (``df``) containing feature values.

    .. function:: compute_and_write_streaming(input: Dict[str, Any], feature_table_name: str, checkpoint_location: Optional[str] = None, trigger: Dict[str, Any] = {'processingTime': '5 minutes'}) -> pyspark.sql.streaming.StreamingQuery

       .. note::

          Experimental: This function may change or be removed in
          a future release without warning.

       Calls the decorated function using the provided input,
       then streams the output :class:`DataFrame <pyspark.sql.DataFrame>` to the
       feature table specified by ``feature_table_name``.

       .. code-block:: python
          :caption: Example:

          compute_customer_features.compute_and_write_streaming(
              input={
                  'data': data,
              },
              feature_table_name='recommender_system.customer_features',
          )

       :param input: If ``input`` is not a dictionary,
          it is passed to the decorated function as the first positional argument.
          If ``input`` is a dictionary, the contents are unpacked and passed
          to the decorated function as keyword arguments.
       :param feature_table_name: A feature table name of the form ``<database_name>.<table_name>``,
          for example ``dev.user_features``.
       :param checkpoint_location: Sets the Structured Streaming ``checkpointLocation`` option.
          By setting a ``checkpoint_location``, Spark Structured Streaming will store progress
          information and intermediate state, enabling recovery after failures.
          This parameter is only supported when the argument ``df``
          is a streaming :class:`DataFrame <pyspark.sql.DataFrame>`.
       :param trigger: ``trigger`` defines the timing of stream data processing,
          the dictionary will be unpacked and passed to :meth:`DataStreamWriter.trigger <pyspark.sql.streaming.DataStreamWriter.trigger>` as arguments.
          For example, ``trigger={'once': True}`` will result in a call to ``DataStreamWriter.trigger(once=True)``.
       :return: A PySpark :class:`StreamingQuery <pyspark.sql.streaming.StreamingQuery>`.
    """
    fs = FeatureStoreClient()

    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    def compute_and_write(
        input: Dict[str, Any], feature_table_name: str, mode: str = MERGE
    ) -> DataFrame:
        """Compute and write to a df."""
        df = func(**input) if isinstance(input, dict) else func(input)
        fs.write_table(
            feature_table_name,
            df,
            mode,
        )
        return df

    def compute_and_write_streaming(
        input: Dict[str, Any],
        feature_table_name: str,
        checkpoint_location: Union[str, None] = None,
        trigger: Dict[str, Any] = FeatureStoreClient._DEFAULT_WRITE_STREAM_TRIGGER,
    ) -> StreamingQuery:
        """Compute and write a streaming df."""
        df = func(**input) if isinstance(input, dict) else func(input)
        if not df.isStreaming:
            raise ValueError(
                "compute_and_write_streaming expects the decorated function to return a streaming DataFrame, "
                "received a non-streaming DataFrame."
            )
        streaming_query = fs.write_table(
            feature_table_name,
            df,
            MERGE,
            checkpoint_location=checkpoint_location,
            trigger=trigger,
        )
        return streaming_query

    wrapper.compute_and_write = compute_and_write
    wrapper.compute_and_write_streaming = compute_and_write_streaming
    return wrapper



"""
Constants defining the online store property names.

These properties are not stored as properties in the Hive metastore,
so do not need to have prefix 'databricks.feature_store'.
"""
TYPE = "type"
HOSTNAME = "hostname"
PORT = "port"
DATABASE_NAME = "database_name"
TABLE_NAME = "table_name"
FEATURES = "features"
DRIVER_NAME = "driver_name"
READ_SECRET_PREFIX = "read_secret_prefix"
WRITE_SECRET_PREFIX = "write_secret_prefix"
REGION = "region"
TTL = "ttl"
ACCOUNT_URI = "account_uri"
CONTAINER_NAME = "container_name"

# Valid property names that are also valid suffixes for user-provided secret prefixes
USER = "user"
PASSWORD = "password"
ACCESS_KEY_ID = "access-key-id"
SECRET_ACCESS_KEY = "secret-access-key"
SESSION_TOKEN = "session-token"
AUTHORIZATION_KEY = "authorization-key"

# List of explicit credential parameters in any OnlineStoreSpec.
# TODO (ML-23105): Remove explicit parameters for MLR 12.0.
EXPLICIT_CREDENTIAL_PARAMS = [
    USER,
    PASSWORD,
    ACCESS_KEY_ID,
    SECRET_ACCESS_KEY,
    SESSION_TOKEN,
]

# Valid Online Store Types
AWS_MYSQL = "aws.rds.mysql"
AWS_AURORA = "aws.rds.aurora"
AZURE_MYSQL = "azure.db.mysql"
AZURE_SQL_SERVER = "azure.db.mssql"
AWS_DYNAMODB = "aws.dynamodb"
AZURE_COSMOSDB = "azure.cosmosdb"

# Valid Publish Auth Types
SECRETS = "secrets"
ROLE = "role"



from abc import ABC, abstractmethod
from datetime import timedelta
from typing import Optional

from databricks.feature_store.constants import (
    OVERWRITE,
    MERGE,
)
from databricks.feature_store.online_store_spec import (
    OnlineStoreSpec,
    AzureSqlServerSpec,
    AzureMySqlSpec,
    AmazonRdsMySqlSpec,
    AmazonDynamoDBSpec,
    AzureCosmosDBSpec,
)

RDBMS_SPECS = [AzureSqlServerSpec, AzureMySqlSpec, AmazonRdsMySqlSpec]
NOSQL_SPECS = [AmazonDynamoDBSpec, AzureCosmosDBSpec]
TTL_SPECS = [AmazonDynamoDBSpec]


def is_rdbms_spec(online_store_spec: OnlineStoreSpec):
    return type(online_store_spec) in RDBMS_SPECS


def is_nosql_spec(online_store_spec: OnlineStoreSpec):
    return type(online_store_spec) in NOSQL_SPECS


def is_ttl_spec(online_store_spec: OnlineStoreSpec):
    """
    Returns if the OnlineStoreSpec supports time to live.
    """
    return type(online_store_spec) in TTL_SPECS


class OnlineTable(ABC):
    def __init__(
        self, name: str, cloud_provider_unique_id: Optional[str], new_table: bool
    ):
        self._name = name
        self._cloud_provider_unique_id = cloud_provider_unique_id
        self._new_table = new_table

    @property
    def name(self):
        return self._name

    @property
    def cloud_provider_unique_id(self):
        return self._cloud_provider_unique_id

    @property
    def is_new_empty_table(self):
        return self._new_table


class OnlineStorePublishClient(ABC):
    def publish(
        self,
        df,
        primary_keys,
        timestamp_keys,
        streaming,
        mode,
        trigger,
        checkpoint_location=None,
        lookback_window: Optional[timedelta] = None,
    ):
        """Write the contents of df to the online store.

        The online store information is obtained from the OnlineStoreSpec
        passed to the derived class constructor. If the online store already contains a table with
        the specified table_name, new data will be merged if mode=MERGE or overwritten
        if mode=OVERWRITE.  mode=OVERWRITE is not supported when streaming=True.

        For more details, see the docstring for
        FeatureStoreClient.publish_table.

        Returns pyspark.sql.streaming.StreamingQuery if streaming=True, else None

        :param lookback_window: This parameter is extracted from the OnlineStoreSpec as part of
          FeatureStoreClient.publish_table, and is only applicable when publishing time series feature tables.
          It defines how far to lookback in the data event time (defined by the timestamp key) from the system time
           when publishing. For example, we publish all data where data event time > system time - lookback window.
        """
        if streaming:
            # The default trigger interval of 5 minutes is arbitrary. We may later run
            # load tests to see if processingTime can be decreased without causing high
            # load on the online store.
            # TODO: Scale test & solicit feedback from the structured streaming team.
            options = {}
            if checkpoint_location is not None:
                options["checkpointLocation"] = checkpoint_location

            # Streaming only supports merge mode, so we call self._publish_merge.
            return (
                df.writeStream.trigger(**trigger)
                .outputMode("update")
                .foreachBatch(
                    lambda mb, _: self._publish_merge(
                        mb, primary_keys, timestamp_keys, lookback_window
                    )
                )
                .options(**options)
                .start()
            )
        # TODO [ML-18325]: move get latest snapshot logic from functions below to client.py
        # Since spark streaming does not natively support get_latest_snapshot logic,
        # we are nesting this logic inside each micro-batch operation. Once this operation
        # is supported by Spark streaming natively in the future, we should extra this logic out
        # to client.py before calling PublishClient.publish().
        elif mode == MERGE:
            self._publish_merge(df, primary_keys, timestamp_keys, lookback_window)
        elif mode == OVERWRITE:
            self._publish_overwrite(df, primary_keys, timestamp_keys, lookback_window)
        else:
            # Should never get here since "mode" should be checked prior to calling
            # this internal API.
            raise ValueError("Invalid publish mode")
        self.close()

    @abstractmethod
    def get_or_create_online_table(self, df, primary_keys, timestamp_keys):
        """
        Create empty online table with expected schema if not exists.
        Returns OnlineTable
        """
        raise NotImplementedError

    @abstractmethod
    def _publish_overwrite(
        self,
        df,
        primary_keys,
        timestamp_keys,
        lookback_window: Optional[timedelta],
    ):
        raise NotImplementedError

    @abstractmethod
    def _publish_merge(
        self,
        df,
        primary_keys,
        timestamp_keys,
        lookback_window: Optional[timedelta],
    ):
        raise NotImplementedError

    @abstractmethod
    def close(self):
        raise NotImplementedError



from databricks.feature_store.utils.rest_utils import http_request, verify_rest_response

_GET_JOB_ENDPOINT = "/api/2.0/jobs/get"


class DatabricksClient:
    """
    This client issues requests to services other than Feature Store.

    If the number of APIs accessed grows too large, we should split this up into per-service APIs
    and import the request/repsonse protos for validation.
    """

    def __init__(self, get_host_creds):
        self._get_host_creds = get_host_creds

    def take_notebook_snapshot(self, notebook_path: str):
        """
        Sends a request to WebApp to take a notebook snapshot, and returns the revision timestamp
        of the snapshot.

        This method should be updated to route the call through the Feature Store service, as
        mlflow does.
        """
        response = http_request(
            self._get_host_creds(),
            "/api/2.0/workspace/take-notebook-snapshot",  # Inlined because this is not a public API
            method="POST",
            json={"hidden": True, "path": notebook_path},
        )
        response = verify_rest_response(
            response, "/api/2.0/workspace/take-notebook-snapshot"
        )
        # It would be better to deserialize the response as a proto, but the Feature Store python
        # wheel should not leak other internal WebApp APIs.
        return response.json()["revision"]["revision_timestamp"]



from databricks.feature_store.api.proto.feature_catalog_pb2 import (
    KeySpec as ProtoKeySpec,
)
from databricks.feature_store.entities._feature_store_object import _FeatureStoreObject


class KeySpec(_FeatureStoreObject):
    """
    Key spec for primary keys, partition keys, and features.
    Encodes name and data type for the feature.
    """

    def __init__(self, name: str, data_type: str, data_type_details: str = None):
        self._name = name
        self._data_type = data_type
        self._data_type_details = data_type_details

    @property
    def name(self):
        return self._name

    @property
    def data_type(self):
        return self._data_type

    @property
    def data_type_details(self):
        return self._data_type_details

    @classmethod
    def from_proto(cls, key_spec_proto):
        return cls(
            name=key_spec_proto.name,
            data_type=key_spec_proto.data_type,
            data_type_details=key_spec_proto.data_type_details,
        )

    def to_proto(self):
        return ProtoKeySpec(
            name=self.name,
            data_type=self.data_type.upper(),
            data_type_details=self.data_type_details,
        )

    def __eq__(self, other):
        if isinstance(other, KeySpec):
            return dict(self) == dict(other)
        return False



from databricks.feature_store.entities.data_type import DataType

OVERWRITE = "overwrite"
MERGE = "merge"
PATH = "path"
TABLE = "table"
CUSTOM = "custom"
PREDICTION_COLUMN_NAME = "prediction"
MODEL_DATA_PATH_ROOT = "feature_store"
UTF8_BYTES_PER_CHAR = 4
MAX_PRIMARY_KEY_STRING_LENGTH_CHARS = 100
MAX_PRIMARY_KEY_STRING_LENGTH_BYTES = (
    MAX_PRIMARY_KEY_STRING_LENGTH_CHARS * UTF8_BYTES_PER_CHAR
)
COMPLEX_DATA_TYPES = [DataType.ARRAY, DataType.MAP]
DATA_TYPES_REQUIRES_DETAILS = COMPLEX_DATA_TYPES + [DataType.DECIMAL]
_DEFAULT_WRITE_STREAM_TRIGGER = {"processingTime": "5 seconds"}
_DEFAULT_PUBLISH_STREAM_TRIGGER = {"processingTime": "5 minutes"}

_WARN = "WARN"
_ERROR = "ERROR"
_SOURCE_FORMAT_DELTA = "delta"



import codecs
import os

import yaml

from mlflow.exceptions import MissingConfigException
from mlflow.utils.file_utils import ENCODING, exists


def write_yaml(root, file_name, data, overwrite=False, sort_keys=True):
    """
    Write dictionary data in yaml format.

    :param root: Directory name.
    :param file_name: Desired file name. Will automatically add .yaml extension if not given
    :param data: data to be dumped as yaml format
    :param overwrite: If True, will overwrite existing files
    """
    if not exists(root):
        raise MissingConfigException("Parent directory '%s' does not exist." % root)

    file_path = os.path.join(root, file_name)
    yaml_file_name = file_path if file_path.endswith(".yaml") else file_path + ".yaml"

    if exists(yaml_file_name) and not overwrite:
        raise Exception("Yaml file '%s' exists as '%s" % (file_path, yaml_file_name))

    try:
        with codecs.open(yaml_file_name, mode="w", encoding=ENCODING) as yaml_file:
            yaml.safe_dump(
                data,
                yaml_file,
                default_flow_style=False,
                allow_unicode=True,
                sort_keys=sort_keys,
            )
    except Exception as e:
        raise e



"""Entities definition for data profiles module."""

from enum import Enum


class ComputationMode(str, Enum):
    MANUAL = "MANUAL"
    AUTO = "AUTO"


class Granularities(str, Enum):
    GLOBAL = "GLOBAL"
    ONE_HOUR = "1 HOUR"
    ONE_DAY = "1 DAY"
    ONE_WEEK = "1 WEEK"
    ONE_MONTH = "1 MONTH"



from databricks.feature_store.entities._feature_store_object import _FeatureStoreObject


class Notebook(_FeatureStoreObject):
    def __init__(
        self,
        notebook_id,
        revision_id,
        notebook_workspace_id,
        feature_table_workspace_id,
    ):
        self._notebook_id = notebook_id
        self._revision_id = revision_id
        self._notebook_workspace_id = notebook_workspace_id
        self._feature_table_workspace_id = feature_table_workspace_id

    @property
    def notebook_id(self):
        return self._notebook_id

    @property
    def revision_id(self):
        return self._revision_id

    @property
    def notebook_workspace_id(self):
        return self._notebook_workspace_id

    @property
    def feature_table_workspace_id(self):
        return self._feature_table_workspace_id

    @classmethod
    def from_proto(cls, notebook_proto):
        return cls(
            notebook_id=notebook_proto.notebook_id,
            revision_id=notebook_proto.revision_id,
            notebook_workspace_id=notebook_proto.notebook_workspace_id,
            feature_table_workspace_id=notebook_proto.feature_table_workspace_id,
        )



from typing import List, Optional

from databricks.feature_store.entities.cloud import Cloud
from databricks.feature_store.entities.store_type import StoreType
from databricks.feature_store.online_store_spec import (
    OnlineStoreSpec,
)
from databricks.feature_store.online_store_spec.online_store_properties import (
    ACCOUNT_URI,
    CONTAINER_NAME,
    AZURE_COSMOSDB,
    AUTHORIZATION_KEY,
    DATABASE_NAME,
    SECRETS,
)


class AzureCosmosDBSpec(OnlineStoreSpec):
    """
    This :class:`OnlineStoreSpec <databricks.feature_store.online_store_spec.online_store_spec.OnlineStoreSpec>`
    implementation is intended for publishing features to Azure Cosmos DB.

    If `database_name` and `container_name` are not provided,
    :meth:`FeatureStoreClient.publish_table <databricks.feature_store.client.FeatureStoreClient.publish_table>`
    will use the offline store's database and table name as the Cosmos DB database and container name.

    The expected read or write secret for Cosmos DB for a given ``{prefix}`` string is
    ``${prefix}-authorization-key``.

    The authorization key can be either the Cosmos DB account primary or secondary key.

    .. note::
        Available in version >= 0.5.0.

    :param account_uri: URI of the Cosmos DB account.
    :param database_name: Database name.
    :param container_name: Container name.
    :param read_secret_prefix: Prefix for read secret.
    :param write_secret_prefix: Prefix for write secret.
    """

    def __init__(
        self,
        *,
        account_uri: str,
        database_name: Optional[str] = None,
        container_name: Optional[str] = None,
        read_secret_prefix: Optional[str] = None,
        write_secret_prefix: str,
    ):
        """Initialize AzureCosmosDBSpec object."""
        super().__init__(
            AZURE_COSMOSDB,
            database_name=database_name,
            read_secret_prefix=read_secret_prefix,
            write_secret_prefix=write_secret_prefix,
            _internal_properties={
                ACCOUNT_URI: account_uri,
                CONTAINER_NAME: container_name,
            },
        )

    @property
    def account_uri(self):
        """
        Account URI of the online store.
        """
        return self._properties[ACCOUNT_URI]

    @property
    def database_name(self):
        """
        Database name.
        """
        return self._properties[DATABASE_NAME]

    @property
    def container_name(self):
        """
        Container name.
        """
        return self._properties[CONTAINER_NAME]

    @property
    def cloud(self):
        """Define the cloud property for the data store."""
        return Cloud.AZURE

    @property
    def store_type(self):
        """Define the data store type."""
        return StoreType.COSMOSDB

    def _lookup_authorization_key_with_write_permissions(self) -> str:
        """
        Authorization key that has write access to the online store, resolved through the write_secret_prefix and dbutils.

        WARNING: do not hold onto the returned secret for longer than necessary, for example saving in
        data structures, files, other persistent backends. Use it only for directly accessing resources
        and then allow the Python VM to remove the reference as soon as it's out of scope.
        """
        return self._lookup_secret_with_write_permissions(AUTHORIZATION_KEY)

    def _validate_credentials(self):
        """
        Validate that the expected credentials were provided and are unambiguous.
        """
        # No validation is currently required since write_secret_prefix is a required parameter.
        pass

    def _valid_secret_suffixes(self) -> List[str]:
        """
        List of valid secret suffixes.
        """
        return [AUTHORIZATION_KEY]

    def _expected_secret_suffixes(self) -> List[str]:
        """
        List of expected secret suffixes.
        """
        return [AUTHORIZATION_KEY]

    def auth_type(self):
        """Publish Auth type."""
        return SECRETS



from databricks.feature_store.online_store_spec.online_store_spec import OnlineStoreSpec
from databricks.feature_store.online_store_spec.amazon_rds_mysql_online_store_spec import (
    AmazonRdsMySqlSpec,
)
from databricks.feature_store.online_store_spec.azure_mysql_online_store_spec import (
    AzureMySqlSpec,
)
from databricks.feature_store.online_store_spec.azure_sql_server_online_store_spec import (
    AzureSqlServerSpec,
)
from databricks.feature_store.online_store_spec.amazon_dynamodb_online_store_spec import (
    AmazonDynamoDBSpec,
)
from databricks.feature_store.online_store_spec.azure_cosmosdb_online_store_spec import (
    AzureCosmosDBSpec,
)

__all__ = [
    "AmazonRdsMySqlSpec",
    "AzureMySqlSpec",
    "AzureSqlServerSpec",
    "AmazonDynamoDBSpec",
    "AzureCosmosDBSpec",
    "OnlineStoreSpec",
]



""" Defines the SqlEngine class, which is used to define a common set of
metadata actions that can be performed on SQL databases which inherit
from this class. These metadata actions differ from traditional writes
on the database, which should be done via the Spark SQL JDBC APIs.
"""

import abc

from contextlib import contextmanager
from pyspark.sql.types import StringType


def format_sql_statement(statement):
    stripped = statement.strip()
    if stripped.endswith(";"):
        return stripped
    return stripped + ";"


def get_string_primary_keys(schema, primary_keys):
    return [
        field.name
        for field in schema.fields
        if field.name in primary_keys and isinstance(field.dataType, StringType)
    ]


def set_primary_keys_not_nullable(schema, primary_keys):
    for field in schema.fields:
        if field.name in primary_keys:
            field.nullable = False


class PublishSqlEngine(abc.ABC):

    INFORMATION_SCHEMA = "INFORMATION_SCHEMA"
    TABLES = "TABLES"
    TABLE_CATALOG = "TABLE_CATALOG"
    TABLE_SCHEMA = "TABLE_SCHEMA"
    TABLE_NAME = "TABLE_NAME"
    COLUMNS = "COLUMNS"
    COLUMN_NAME = "COLUMN_NAME"
    DATA_TYPE = "DATA_TYPE"
    NUMERIC_PRECISION = "NUMERIC_PRECISION"
    NUMERIC_SCALE = "NUMERIC_SCALE"

    @abc.abstractmethod
    def __init__(self, online_store, spark_session):
        self.online_store = online_store
        self.spark_session = spark_session

        # This member variable should not be accessed directly. Instead, use
        # self._sql_connection.
        self._cached_sql_connection = None

    def _get_connection(self):
        jvm = self.spark_session.sparkContext._jvm
        if not jvm:
            from pyspark.java_gateway import launch_gateway

            jvm = launch_gateway().jvm
        return jvm.java.sql.DriverManager.getConnection(
            self.jdbc_url,
            self.online_store._lookup_jdbc_auth_user_with_write_permissions(),
            self.online_store._lookup_password_with_write_permissions(),
        )

    @property
    def _sql_connection(self):
        """
        A connection to the SQL database that uses a Java driver.
        We may later update this to use a Python database connector.
        """
        if self._cached_sql_connection:
            return self._cached_sql_connection
        self._cached_sql_connection = self._get_connection()
        return self._cached_sql_connection

    def close(self):
        if self._cached_sql_connection:
            self._sql_connection.close()
            self._cached_sql_connection = None

    def get_online_tables(self, table_names=[]):
        raise NotImplementedError

    def get_column_types(self, table_name):
        raise NotImplementedError

    def add_columns(self, table_name, columns):
        raise NotImplementedError

    def create_table_like(self, table_name, like_table_name, primary_keys):
        raise NotImplementedError

    def rename_table(self, from_name, to_name):
        raise NotImplementedError

    def merge_table_into(self, dst_table_name, src_table_name, columns, primary_keys):
        raise NotImplementedError

    def create_empty_table(
        self, table_name, schema, jdbc_url, connection_properties, primary_keys
    ):
        raise NotImplementedError

    def drop_table(self, table_name):
        self._run_sql_update(f"DROP TABLE IF EXISTS {self._sql_safe_name(table_name)}")

    def add_primary_keys(self, table_name, primary_keys):
        self._run_sql_update(
            f'ALTER TABLE {self._sql_safe_name(table_name)} ADD PRIMARY KEY({", ".join(f"{self._sql_safe_name(pk)}" for pk in primary_keys)})',
        )

    @classmethod
    def _sql_safe_name(cls, name):
        raise NotImplementedError

    def _make_empty_df(self, schema):
        return self.spark_session.createDataFrame([], schema)

    @property
    def jdbc_url(self):
        raise NotImplementedError

    @property
    def jdbc_properties(self):
        raise NotImplementedError

    @contextmanager
    def in_transaction(self):
        self._sql_connection.setAutoCommit(False)  # Start transaction
        try:
            yield
            self._sql_connection.commit()  # End transaction
        except:
            self._sql_connection.rollback()
            raise
        finally:
            self._sql_connection.setAutoCommit(True)

    def _run_sql_update(self, statement):
        stmt = self._sql_connection.createStatement()
        stmt.executeUpdate(format_sql_statement(statement))

    def _run_sql_query(self, query):
        stmt = self._sql_connection.createStatement()
        return stmt.executeQuery(format_sql_statement(query))



import logging
from typing import Dict
from databricks.feature_store.api.proto.feature_catalog_pb2 import (
    FeatureTable as ProtoFeatureTable,
)
from databricks.feature_store.entities._feature_store_object import _FeatureStoreObject
from mlflow.utils.annotations import deprecated


class FeatureTable(_FeatureStoreObject):
    """
    Value class describing one feature table.

    This will typically not be instantiated directly, instead the
    :meth:`FeatureStoreClient.create_table <databricks.feature_store.client.FeatureStoreClient.create_table>`
    will create :class:`.FeatureTable` objects.
    """

    def __init__(
        self,
        name,
        table_id,
        description,
        primary_keys,
        partition_columns,
        features,
        creation_timestamp=None,
        online_stores=[],
        notebook_producers=[],
        job_producers=[],
        table_data_sources=[],
        path_data_sources=[],
        custom_data_sources=[],
        timestamp_keys=[],
        tags=None,
    ):

        """Initialize a FeatureTable object."""
        self.name = name
        self.table_id = table_id
        self.description = description
        self.primary_keys = primary_keys
        self.partition_columns = partition_columns
        self.features = features
        self.creation_timestamp = creation_timestamp
        self.online_stores = online_stores
        self.notebook_producers = notebook_producers
        self.job_producers = job_producers
        self.table_data_sources = table_data_sources
        self.path_data_sources = path_data_sources
        self.custom_data_sources = custom_data_sources
        self.timestamp_keys = timestamp_keys
        self._tags = tags

    @property
    @deprecated("FeatureTable.primary_keys", since="v0.3.6")
    def keys(self):
        return self.primary_keys

    @property
    def tags(self) -> Dict[str, str]:
        """
        Get the tags associated with the feature table.

        :return a Dictionary of all tags associated with the feature table as key/value pairs
        """
        if self._tags is None:
            # If no tags are set, self._tags is expected an empty dictionary.
            raise ValueError(
                "Internal error: tags have not been fetched for this FeatureTable instance"
            )
        return self._tags

    @classmethod
    def from_proto(cls, feature_table_proto: ProtoFeatureTable):
        """Return a FeatureStore object from a proto.

        Note: `repeated` proto fields are cast from
              `google.protobuf.pyext._message.RepeatedScalarContainer`
              to list.

        :param FeatureTable feature_table_proto: Prototype for a :class:`.FeatureTable` object.
        :return FeatureTable: a FeatureStore object from a proto.
        """
        return cls(
            name=feature_table_proto.name,
            table_id=feature_table_proto.id,
            description=feature_table_proto.description,
            primary_keys=list(feature_table_proto.primary_keys),
            partition_columns=list(feature_table_proto.partition_keys),
            features=list(feature_table_proto.features),
            creation_timestamp=feature_table_proto.creation_timestamp,
            online_stores=list(feature_table_proto.online_stores),
            notebook_producers=list(feature_table_proto.notebook_producers),
            job_producers=list(feature_table_proto.job_producers),
            table_data_sources=[
                s.table for s in feature_table_proto.data_sources if s.table
            ],
            path_data_sources=[
                s.path for s in feature_table_proto.data_sources if s.path
            ],
            custom_data_sources=[
                s.custom_source
                for s in feature_table_proto.data_sources
                if s.custom_source
            ],
            timestamp_keys=list(feature_table_proto.timestamp_keys),
            tags=None,
        )



import logging
from databricks.feature_store.utils import validation_utils
from databricks.feature_store.hive_client import HiveClient

_logger = logging.getLogger(__name__)


class HiveClientHelper:
    """
    Helper functions that wrap calls to the hive client with application specific business logic.
    """

    def __init__(self, hive_client: HiveClient):
        self._hive_client = hive_client

    def check_database_exists(self, feature_table_name):
        """
        Check for the existence of a database.

        feature_table_name should have the form <database_name>.<table_name>.
        Check whether database_name is a database in the metastore.
        """
        database_name, _ = feature_table_name.split(".")
        if not self._hive_client.database_exists(database_name):
            raise ValueError(
                f"Database '{database_name}' does not exist in the Hive metastore."
            )

    def check_feature_table_exists_in_hive(self, feature_table_name):
        validation_utils.check_qualified_name(feature_table_name)
        self.check_database_exists(feature_table_name)
        if not self._hive_client.table_exists(feature_table_name):
            raise ValueError(
                f"The feature data table for '{feature_table_name}' could not be found."
            )



from dateutil import parser, tz
import os
from typing import Any, Dict, List

import mlflow
from mlflow.store.artifact.artifact_repository_registry import get_artifact_repository
from mlflow.store.artifact.models_artifact_repo import ModelsArtifactRepository
from mlflow.store.artifact.runs_artifact_repo import RunsArtifactRepository
from mlflow.exceptions import RestException
from mlflow.utils import databricks_utils
from databricks.feature_store.entities.store_type import StoreType
from databricks.feature_store.entities.cloud import Cloud

from databricks.feature_store.constants import MODEL_DATA_PATH_ROOT


def as_list(obj, default=None):
    if not obj:
        return default
    elif isinstance(obj, list):
        return obj
    else:
        return [obj]


def as_directory(path):
    if path.endswith("/"):
        return path
    return f"{path}/"


def is_empty(target: str):
    return target is None or len(target.strip()) == 0


class _NoDbutilsError(Exception):
    pass


def _get_dbutils():
    try:
        import IPython

        ip_shell = IPython.get_ipython()
        if ip_shell is None:
            raise _NoDbutilsError
        return ip_shell.ns_table["user_global"]["dbutils"]
    except ImportError:
        raise _NoDbutilsError
    except KeyError:
        raise _NoDbutilsError


def get_canonical_online_store_name(online_store):
    if online_store.cloud == Cloud.AWS:
        if online_store.store_type == StoreType.MYSQL:
            return "Amazon RDS MySQL"
        elif online_store.store_type == StoreType.AURORA_MYSQL:
            return "Amazon Aurora (MySQL-compatible)"
        elif online_store.store_type == StoreType.DYNAMODB:
            return "Amazon DynamoDB"
    elif online_store.cloud == Cloud.AZURE:
        if online_store.store_type == StoreType.MYSQL:
            return "Azure Database for MySQL"
        elif online_store.store_type == StoreType.SQL_SERVER:
            return "Azure SQL Database"


# TODO (ML-23163): Consider using `datetime.fromisoformat` when Python 3.7+ is available
# Python's datetime.fromisoformat` is only supported in 3.7+, but the Runbot is currently on 3.6
def utc_timestamp_ms_from_iso_datetime_string(date_string):
    # Python uses seconds for its time granularity, so we multiply by 1000 to convert to milliseconds.
    # The Feature Store backend returns timestamps in milliseconds, so this allows for direct comparisons.
    dt = parser.parse(date_string)
    utc_dt = dt.replace(tzinfo=tz.tzutc())
    return 1000 * utc_dt.timestamp()


def pip_depependency_pinned_major_version(pip_package_name, major_version):
    """
    Generate a pip dependency string that is pinned to a major version, for example: "databricks-feature-lookup==0.*"
    """
    return f"{pip_package_name}=={major_version}.*"


def add_mlflow_pip_depependency(conda_env, pip_package_name):
    """
    Add a new pip dependency to the conda environment taken from the raw MLflow model.  This method should only be
    called for conda environments created by MLflow rather than for generic conda environments, because it assumes
    the conda environment already contains pip as a dependency.  In the case of MLflow models, this is a safe
    assumption because MLflow always needs to add "mlflow" to the conda environment's pip dependencies.

    This is idempotent and will not add a pip package that is already present in the list of pip packages.
    """
    if pip_package_name is None or len(pip_package_name) == 0:
        raise ValueError(
            "Unexpected input: missing or empty pip_package_name parameter"
        )

    found_pip_dependency = False
    if conda_env is not None:
        for dep in conda_env["dependencies"]:
            if isinstance(dep, dict) and "pip" in dep:
                found_pip_dependency = True
                pip_deps = dep["pip"]
                if pip_package_name not in pip_deps:
                    pip_deps.append(pip_package_name)
        # Fail early rather than at model inference time
        if "dependencies" in conda_env and not found_pip_dependency:
            raise ValueError(
                "Unexpected input: mlflow conda_env did not contain pip as a dependency"
            )


def download_model_artifacts(model_uri, dir):
    """
    Downloads model artifacts from model_uri to dir.

    :param model_uri: The location, in URI format, of a model. Must be either in the model registry
      (``models:/<model_name>/<model_version>``, ``models:/<model_name>/<stage>``) or the MLflow
      artifact store (``runs:/<mlflow_run_id>/run-relative/path/to/model``).
    :param dir: Location to place downloaded model artifacts.
    """
    # Check model URI format
    if not model_uri:
        raise ValueError("Missing model URI.")
    if not ModelsArtifactRepository.is_models_uri(
        model_uri
    ) and not RunsArtifactRepository.is_runs_uri(model_uri):
        raise ValueError(
            f"Invalid model URI '{model_uri}'."
            f"Use ``models:/model_name>/<version_number>`` or "
            f"``runs:/<mlflow_run_id>/run-relative/path/to/model``."
        )

    try:
        repo = get_artifact_repository(model_uri)
    except RestException as e:
        raise ValueError(f"The model at '{model_uri}' does not exist.", e)

    artifact_path = os.path.join(mlflow.pyfunc.DATA, MODEL_DATA_PATH_ROOT)
    if len(repo.list_artifacts(artifact_path)) == 0:
        raise ValueError(
            f"No suitable model found at '{model_uri}'. Either no model exists in this "
            f"artifact location or an existing model was not packaged with Feature Store metadata. "
            f"Only models logged by FeatureStoreClient.log_model can be used in inference."
        )

    return repo.download_artifacts(artifact_path="", dst_path=dir)


def validate_params_non_empty(params: Dict[str, Any], expected_params: List[str]):
    """
    Validate that none of the expected parameters are empty, otherwise raise a Value error
    for the first encountered empty parameter.

    Tested with the following param types:

    - str
    - Dict
    - List

    :param params: A dictionary of param names -> param values, for example as returned by locals()
    :param expected_params: List of params to check as non_empty
    """
    for expected_param in expected_params:
        if expected_param not in params:
            raise ValueError(
                f'Internal error: expected parameter "{expected_param}" not found in params dictionary'
            )
        param_value = params[expected_param]
        if not param_value:
            raise ValueError(f'Parameter "{expected_param}" cannot be empty')


def is_in_databricks_job():
    """
    Overrides the behavior of the mlflow databricks_utils.is_in_databricks_job() to account for the fact that
    some jobs have job_id but no run_id, for example one-time job runs.
    """
    try:
        return databricks_utils.get_job_id() is not None
    except Exception:
        return False


def is_in_databricks_env():
    """
    Determine if we are running in a Databricks environment (DBR, MLR, DLT, DCS, Mlflow Projects, Run Cmd 1.2 API, etc)

    If any invoked methods raise an exception, swallow the exception and return False out of an abundance of caution.
    """
    try:
        return (
            is_in_databricks_job()
            or databricks_utils.is_in_databricks_notebook()
            or databricks_utils.is_in_databricks_runtime()
        )
    except Exception:
        return False



import abc
from typing import List, Optional
from urllib.parse import urlparse

from azure.cosmos import CosmosClient
from azure.cosmos.exceptions import CosmosResourceNotFoundError
from pyspark.sql import SparkSession
from pyspark.sql.types import StructType, TimestampType, DateType

from databricks.feature_store.online_store_spec import AzureCosmosDBSpec
from databricks.feature_store.utils.cosmosdb_utils import (
    generate_cosmosdb_safe_data_types,
    generate_cosmosdb_primary_key,
    generate_cosmosdb_id,
    SERVERLESS_TABLE_PROPERTIES,
    PROVISIONED_TABLE_PROPERTIES,
    is_serverless_throughput_offer_error,
    create_table_statement,
    create_database_statement,
)


class PublishCosmosDBEngine(abc.ABC):
    def __init__(
        self, online_store_spec: AzureCosmosDBSpec, spark_session: SparkSession
    ):
        self.spark_session = spark_session
        self.account_uri = online_store_spec.account_uri
        self.authorization_key = (
            online_store_spec._lookup_authorization_key_with_write_permissions()
        )
        self.database_name = online_store_spec.database_name
        self.container_name = online_store_spec.container_name
        self._initialize_cosmosdb_spark_connector()

    def _initialize_cosmosdb_spark_connector(self):
        """
        Initialize and set the configs required to create and write to Cosmos DB containers.
        """
        # Configurations used by the Spark writer to write to Cosmos DB. See example usage and config documentation:
        # https://github.com/Azure/azure-sdk-for-java/blob/main/sdk/cosmos/azure-cosmos-spark_3_2-12/docs/quick-start.md
        # https://github.com/Azure/azure-sdk-for-java/blob/main/sdk/cosmos/azure-cosmos-spark_3_2-12/docs/configuration-reference.md
        self.cosmosdb_spark_writer_options = {
            "spark.cosmos.accountEndpoint": self.account_uri,
            "spark.cosmos.accountKey": self.authorization_key,
            "spark.cosmos.database": self.database_name,
            "spark.cosmos.container": self.container_name,
            # We explicitly enforce the default value "ItemOverwrite" to guarantee upsert behavior.
            "spark.cosmos.write.strategy": "ItemOverwrite",
        }

        # Configurations used when creating databases and containers through the Catalog API.
        # https://github.com/Azure/azure-sdk-for-java/blob/main/sdk/cosmos/azure-cosmos-spark_3_2-12/docs/catalog-api.md
        self.spark_session.conf.set(
            "spark.sql.catalog.cosmosCatalog", "com.azure.cosmos.spark.CosmosCatalog"
        )
        self.spark_session.conf.set(
            "spark.sql.catalog.cosmosCatalog.spark.cosmos.accountEndpoint",
            self.account_uri,
        )
        self.spark_session.conf.set(
            "spark.sql.catalog.cosmosCatalog.spark.cosmos.accountKey",
            self.authorization_key,
        )

    def _validate_timestamp_keys(self, schema: StructType, timestamp_keys: List[str]):
        # Check: We only support a single timestamp key in Cosmos DB.
        # TODO: ML-19665 add similar check in backend RPC validation.
        # TODO (ML-22021): move this validation to FeatureStoreClient.publish
        if len(timestamp_keys) > 1:
            raise ValueError(
                "Only one timestamp key is supported in Cosmos DB online store."
            )

        # The backend validates that the timestamp key is of Date, Timestamp type. The validation is duplicated
        # here in this client version to allow for the introduction of integer timestamp keys in future clients,
        # as the current Cosmos DB data model is not compatible with integer timestamp keys.
        if timestamp_keys and not type(schema[timestamp_keys[0]].dataType) in [
            DateType,
            TimestampType,
        ]:
            raise ValueError(
                "The timestamp key for Cosmos DB must be of either Date or Timestamp type."
            )

    def close(self):
        """
        Performs any close operations on the Cosmos DB connections. Cosmos DB connections are
        stateless http connections and hence does not need to be closed.
        :return:
        """
        pass

    def database_and_container_exist(self) -> bool:
        """
        Cosmos DB client is conservative in validation and does not check that (database, container) exists.
        Calling _get_properties forces the validation that both exist.

        The Spark connector can't check if the database and container exist, so the Python SDK is used here.
        """
        try:
            container_client = (
                CosmosClient(self.account_uri, self.authorization_key)
                .get_database_client(self.database_name)
                .get_container_client(self.container_name)
            )
            container_client._get_properties()
            return True
        except CosmosResourceNotFoundError:
            return False

    def get_cloud_provider_unique_id(self) -> Optional[str]:
        """
        Generate the expected container URI: https://{databaseaccount}.documents.azure.com/dbs/{db}/colls/{coll}
        https://docs.microsoft.com/en-us/rest/api/cosmos-db/cosmosdb-resource-uri-syntax-for-rest
        Returns the container URI if both the database and container exist, otherwise None.
        """
        if self.database_and_container_exist():
            # If the database and container exist, the account URI is expected to be well formed as below:
            # https://{databaseaccount}.documents.azure.com:443/, where port 443 and path "/" are optional.
            # Extract {databaseaccount}.documents.azure.com, dropping port and path if present.
            base_uri = urlparse(self.account_uri).hostname

            # Cosmos DB expectes the database account name to be lowercase, so we force that here to avoid generating
            # an unexpected container URI as the CosmosClient treats the URI in a case-insensitive manner.
            # https://docs.microsoft.com/en-us/azure/cosmos-db/sql/create-cosmosdb-resources-portal

            # Database and container name are case sensitive and should be used as-is
            return f"https://{base_uri.lower()}/dbs/{self.database_name}/colls/{self.container_name}"
        else:
            return None

    def create_empty_table(
        self,
        schema: StructType,
        primary_keys: List[str],
        timestamp_keys: List[str],
    ) -> str:
        """
        This method should handle creating the correct online table schema and state.
        (e.g. combining PKs, dropping the timestamp key, enabling ttl)
        """
        self._validate_timestamp_keys(schema, timestamp_keys)

        # All validations should be done prior to database or container creation.
        # If this method is called, then either the database or container did not exist.
        self.spark_session.sql(create_database_statement(self.database_name))

        create_provisioned_table_error = None
        try:
            self.spark_session.sql(
                create_table_statement(
                    self.database_name,
                    self.container_name,
                    PROVISIONED_TABLE_PROPERTIES,
                )
            )
        except Exception as e:
            # Store the exception and conditionally handle it in the finally block to avoid raising nested exceptions.
            # Nested exceptions are surfaced in order they were raised, which results in: 1. a verbose traceback,
            # 2. the Databricks Notebook UI displaying only the first exception message in the cell output.
            create_provisioned_table_error = e
        finally:
            # Creating containers with throughput offers (used for provisioned accounts) throws for serverless accounts.
            # If an exception is thrown for this reason, attempt to create a container without throughput.
            # Otherwise, re-raise the original error.
            if create_provisioned_table_error:
                if is_serverless_throughput_offer_error(create_provisioned_table_error):
                    self.spark_session.sql(
                        create_table_statement(
                            self.database_name,
                            self.container_name,
                            SERVERLESS_TABLE_PROPERTIES,
                        )
                    )
                else:
                    raise create_provisioned_table_error

        # re-fetch the container URI now that the container is guaranteed to exist.
        return self.get_cloud_provider_unique_id()

    def write_table(
        self,
        df,
        schema: StructType,
        primary_keys: List[str],
        timestamp_keys: List[str],
    ):
        """
        This method should handle writing the correct online table schema.
        (e.g. combining PKs, writing online-safe data types, generating ttl)
        """
        self._validate_timestamp_keys(schema, timestamp_keys)

        # Write the DataFrame to Cosmos DB. All validations should be done prior.
        # Convert df to online-safe data types, generate the concatenated primary key and id columns
        df = generate_cosmosdb_safe_data_types(df)
        df = generate_cosmosdb_primary_key(df, primary_keys)

        # Window mode publish should use the timestamp key as the id property to enable range lookups.
        ts_col = timestamp_keys[0] if self.is_timeseries_window_publish else None
        df = generate_cosmosdb_id(df, column_for_id=ts_col)

        df.write.format("cosmos.oltp").options(
            **self.cosmosdb_spark_writer_options
        ).mode("APPEND").save()

    def generate_df_with_ttl_if_required(self, df, timestamp_keys: List[str]):
        """
        Convert the timestamp column to a TTL column expected by Cosmos DB.

        Currently is a no-op as TTL is not supported.
        """
        return df

    @property
    def is_timeseries_window_publish(self):
        """
        Determine if this publish engine will publish a window of a time series dataframe.
        This is currently equivalent to if TTL is defined in the online store spec.

        Currently returns False as TTL is not supported.
        """
        return False

    @property
    def table_name(self) -> str:
        """Table equivalent of this publish engine."""
        return self.container_name



from databricks.feature_store.online_store_spec import (
    OnlineStoreSpec,
)
from databricks.feature_store.online_store_spec.online_store_properties import (
    HOSTNAME,
    USER,
    AZURE_MYSQL,
    PORT,
    DATABASE_NAME,
    SECRETS,
)
from databricks.feature_store.entities.store_type import StoreType
from databricks.feature_store.entities.cloud import Cloud
from typing import Union


class AzureMySqlSpec(OnlineStoreSpec):
    """
    Define the `AzureMySqlSpec` class.

    This :class:`OnlineStoreSpec <databricks.feature_store.online_store_spec.online_store_spec.OnlineStoreSpec>`
    implementation is intended for publishing
    features to Azure Database for MySQL.

    See :class:`OnlineStoreSpec  <databricks.feature_store.online_store_spec.online_store_spec.OnlineStoreSpec>`
    documentation for more usage information,
    including parameter descriptions.

    :param hostname: Hostname to access online store.
    :param port: Port number to access online store.
    :param user: Username that has access to the online store. **Deprecated** as of version 0.6.0.
      Use ``write_secret_prefix`` instead.
    :param password: Password to access the online store. **Deprecated** as of version 0.6.0.
      Use ``write_secret_prefix`` instead.
    :param database_name: Database name.
    :param table_name: Table name.
    :param driver_name: Name of custom JDBC driver to access the online store.
    :param read_secret_prefix: Prefix for read secret.
    :param write_secret_prefix: Prefix for write secret.
    """

    # TODO [ML-15546]: Identify clear mechanism to inherit constructor pydocs from base class and
    #   remove `See xxx documentation for more usage information` section.
    # TODO (ML-23105): Remove explicit parameters for MLR 12.0.
    def __init__(
        self,
        hostname: str,
        port: int,
        user: Union[str, None] = None,
        password: Union[str, None] = None,
        database_name: Union[str, None] = None,
        table_name: Union[str, None] = None,
        driver_name: Union[str, None] = None,
        read_secret_prefix: Union[str, None] = None,
        write_secret_prefix: Union[str, None] = None,
    ):
        """Initialize AzureMySqlSpec objects."""
        super().__init__(
            AZURE_MYSQL,
            hostname,
            port,
            user,
            password,
            database_name=database_name,
            table_name=table_name,
            driver_name=driver_name,
            read_secret_prefix=read_secret_prefix,
            write_secret_prefix=write_secret_prefix,
        )

    @property
    def hostname(self):
        """Hostname to access the online store."""
        return self._properties[HOSTNAME]

    @property
    def port(self):
        """Port number to access the online store."""
        return self._properties[PORT]

    @property
    def database_name(self):
        """Database name."""
        return self._properties[DATABASE_NAME]

    @property
    def cloud(self):
        """Define the cloud the fature store runs."""
        return Cloud.AZURE

    @property
    def store_type(self):
        """Define the data store type."""
        return StoreType.MYSQL

    def auth_type(self):
        """Publish Auth type."""
        return SECRETS

    def _lookup_jdbc_auth_user_with_write_permissions(self):
        """
        Generate the user authentication string used in the JDBC URL.

        On AzureMysql it requires the format: "user@server"
        """
        # https://docs.microsoft.com/en-us/azure/mysql/howto-connection-string
        # servername is extracted from hostname (<servername>.mysql.database.azure.com)
        server_name = self._properties[HOSTNAME].split(".")[0]

        return f"{self._lookup_user_with_write_permissions()}@{server_name}"



import copy

from databricks.feature_store.constants import MAX_PRIMARY_KEY_STRING_LENGTH_CHARS
from databricks.feature_store.publish_engine.publish_sql_engine import (
    PublishSqlEngine,
    get_string_primary_keys,
    set_primary_keys_not_nullable,
)
from databricks.feature_store.utils.spark_utils import (
    get_columns_of_type_short,
    serialize_complex_data_types,
)


class PublishMySqlEngine(PublishSqlEngine):
    CHARSET = "utf8mb4"
    COLLATION = "utf8mb4_bin"
    PRIMARY_KEY_STRING_ERROR = "Data too long for column"
    MAX_VARCHAR_LENGTH_CHARS = 1000

    def __init__(self, online_store, spark_session):
        super().__init__(online_store, spark_session)

    def get_online_tables(self, table_names=[]):
        query = (
            f"SELECT {PublishSqlEngine.TABLE_NAME} FROM {PublishSqlEngine.INFORMATION_SCHEMA}.{PublishSqlEngine.TABLES} "
            f"WHERE {PublishSqlEngine.TABLE_SCHEMA}='{self.online_store.database_name}'"
        )
        if len(table_names) > 0:
            table_names_str = ", ".join(
                [f"'{table_name}'" for table_name in table_names]
            )
            query += f" AND {PublishSqlEngine.TABLE_NAME} IN ({table_names_str})"
        return self._run_sql_query(query)

    def get_column_types(self, table_name):
        query = (
            f"SELECT {PublishSqlEngine.COLUMN_NAME}, {PublishSqlEngine.DATA_TYPE}, {PublishSqlEngine.NUMERIC_PRECISION}, {PublishSqlEngine.NUMERIC_SCALE} "
            f"FROM {PublishSqlEngine.INFORMATION_SCHEMA}.{PublishSqlEngine.COLUMNS} "
            f"WHERE {PublishSqlEngine.TABLE_SCHEMA}='{self.online_store.database_name}' "
            f"AND {PublishSqlEngine.TABLE_NAME}='{table_name}'"
        )
        return self._run_sql_query(query)

    def add_columns(self, table_name, columns):
        if len(columns) == 0:
            return
        new_columns_str = ", ".join(
            [f"{self._sql_safe_name(name)} {type}" for (name, type) in columns]
        )
        self._run_sql_update(
            f"ALTER TABLE {self._sql_safe_name(table_name)} ADD COLUMN ({new_columns_str})"
        )

    def create_table_like(self, table_name, like_table_name, primary_keys):
        self._run_sql_update(
            f"CREATE TABLE IF NOT EXISTS {self._sql_safe_name(table_name)} LIKE {self._sql_safe_name(like_table_name)}",
        )

    def rename_table(self, from_name, to_name):
        self._run_sql_update(
            f"ALTER TABLE {self._sql_safe_name(from_name)} RENAME TO {self._sql_safe_name(to_name)}"
        )

    def merge_table_into(self, dst_table_name, src_table_name, columns, primary_keys):
        columns_str = ", ".join(f"{self._sql_safe_name(c)}" for c in columns)
        update_columns = [col for col in columns if col not in primary_keys]
        update_str = ", ".join(
            [
                f"{self._sql_safe_name(feat)}={self._sql_safe_name(src_table_name)}.{self._sql_safe_name(feat)}"
                for feat in update_columns
            ]
        )

        update_cmd = (
            f"INSERT INTO {self._sql_safe_name(dst_table_name)} ({columns_str}) "
            f"SELECT {columns_str} FROM {self._sql_safe_name(src_table_name)} "
            f"ON DUPLICATE KEY UPDATE {update_str}"
        )
        self._run_sql_update(update_cmd)

    def _update_create_table_column_types_option(
        self, writer, string_pk_cols, short_cols
    ):
        string_pk_col_types = [
            f"{col} VARCHAR({MAX_PRIMARY_KEY_STRING_LENGTH_CHARS})"
            for col in string_pk_cols
        ]
        # Explicitly specify the SHORT type to prevent upcasting to INTEGER by the Spark writer
        # SHORT is specified rather than MySql's SMALLINT as the documentation asks for a Spark Sql type
        # However, SHORT and SMALLINT are equivalent. SHORT will be written as SMALLINT in MySql
        # https://spark.apache.org/docs/latest/sql-data-sources-jdbc.html
        short_col_types = [f"{col} SHORT" for col in short_cols]

        all_col_types = string_pk_col_types + short_col_types
        if len(all_col_types) > 0:
            writer = writer.option("createTableColumnTypes", ", ".join(all_col_types))
        return writer

    def create_empty_table(
        self, table_name, schema, jdbc_url, connection_properties, primary_keys
    ):
        schema = copy.deepcopy(schema)
        set_primary_keys_not_nullable(schema, primary_keys)
        empty_df = self._make_empty_df(schema)

        string_pk_cols = get_string_primary_keys(schema, primary_keys)
        short_cols = get_columns_of_type_short(schema)

        serialized_df = serialize_complex_data_types(empty_df)
        writer = self._update_create_table_column_types_option(
            serialized_df.write, string_pk_cols, short_cols
        )
        writer.option(
            "createTableOptions",
            f"DEFAULT CHARSET={self.CHARSET} COLLATE={self.COLLATION}",
        ).jdbc(
            url=jdbc_url,
            table=table_name,
            mode="overwrite",
            properties=connection_properties,
        )
        self.add_primary_keys(table_name, primary_keys)

    @classmethod
    def _sql_safe_name(cls, name):
        # MySQL requires `xxx` format to safely handle identifiers that contain special characters or are reserved words.
        return f"`{name}`"

    @property
    def jdbc_url(self):
        return f"jdbc:mysql://{self.online_store.hostname}:{self.online_store.port}/{self.online_store.database_name}"

    @property
    def jdbc_properties(self):
        properties = {
            "user": self.online_store._lookup_jdbc_auth_user_with_write_permissions(),
            "password": self.online_store._lookup_password_with_write_permissions(),
        }
        if self.online_store.driver:
            properties["driver"] = self.online_store.driver
        return properties



from typing import Optional

from databricks.feature_store.entities._feature_store_object import _FeatureStoreObject
from databricks.feature_store.entities.cloud import Cloud
from databricks.feature_store.entities.online_store_metadata import (
    OnlineStoreAdditionalMetadata,
    MySqlMetadata,
    SqlServerMetadata,
    DynamoDbMetadata,
    CosmosDbMetadata,
)
from databricks.feature_store.entities.store_type import StoreType
from databricks.feature_store.protos.feature_store_serving_pb2 import (
    OnlineStoreDetailed as ProtoOnlineStoreDetailed,
)


class OnlineStoreDetailed(_FeatureStoreObject):
    """
    This class corresponds to the OnlineStoreDetailed proto message and describes an online store. It was introduced
    in v0.3.8, so previously deprecated fields (host, port) are unnecessary.
    """

    def __init__(
        self,
        cloud: Cloud,
        store_type: StoreType,
        name: str,
        creation_timestamp: int,
        last_updated_timestamp: int,
        creator_id: str,
        last_update_user_id: str,
        read_secret_prefix: Optional[str],
        additional_metadata: OnlineStoreAdditionalMetadata,
    ):

        """Initialize a OnlineStoreDetailed object."""
        self.cloud = cloud
        self.store_type = store_type
        self.name = name
        self.creation_timestamp = creation_timestamp
        self.last_updated_timestamp = last_updated_timestamp
        self.creator_id = creator_id
        self.last_update_user_id = last_update_user_id
        self.read_secret_prefix = read_secret_prefix
        self.additional_metadata = additional_metadata

    @classmethod
    def from_proto(cls, the_proto: ProtoOnlineStoreDetailed):
        """
        Return an OnlineStoreDetailed object from an OnlineStoreDetailed proto.

        Note: `repeated` proto fields are cast from `google.protobuf.pyext._message.RepeatedScalarContainer` to list.
        Additionally, relevant fields (e.g. additional_metadata) are converted to Python representations.
        """
        additional_metadata = None
        if the_proto.WhichOneof("additional_metadata") == "mysql_metadata":
            additional_metadata = MySqlMetadata.from_proto(the_proto.mysql_metadata)
        elif the_proto.WhichOneof("additional_metadata") == "sql_server_metadata":
            additional_metadata = SqlServerMetadata.from_proto(
                the_proto.sql_server_metadata
            )
        elif the_proto.WhichOneof("additional_metadata") == "dynamodb_metadata":
            additional_metadata = DynamoDbMetadata.from_proto(
                the_proto.dynamodb_metadata
            )
        elif the_proto.WhichOneof("additional_metadata") == "cosmosdb_metadata":
            additional_metadata = CosmosDbMetadata.from_proto(
                the_proto.cosmosdb_metadata
            )
        else:
            raise ValueError("Unsupported Store Type: " + str(the_proto))
        return cls(
            cloud=the_proto.cloud,
            store_type=the_proto.store_type,
            name=the_proto.name,
            creation_timestamp=the_proto.creation_timestamp,
            last_updated_timestamp=the_proto.last_updated_timestamp,
            creator_id=the_proto.creator_id,
            last_update_user_id=the_proto.last_update_user_id,
            read_secret_prefix=the_proto.read_secret_prefix,
            additional_metadata=additional_metadata,
        )



import sys
import uuid
from mlflow.utils._spark_utils import _get_active_spark_session
from mlflow.utils import databricks_utils
from py4j.java_gateway import CallbackServerParameters
from pyspark import SparkContext

import logging

_JAVA_PACKAGE = "org.mlflow.spark.autologging"

_logger = logging.getLogger(__name__)


def _get_spark_major_version(sc):
    spark_version_parts = sc.version.split(".")
    spark_major_version = None
    if len(spark_version_parts) > 0:
        spark_major_version = int(spark_version_parts[0])
    return spark_major_version


def _get_jvm_event_publisher():
    """
    Get JVM-side object implementing the following methods:
    - init() for initializing JVM state needed for attaching a SparkListener to watch for datasource
    - register(subscriber) for registering subscribers to receive datasource events
    """
    jvm = SparkContext._gateway.jvm
    qualified_classname = "{}.{}".format(_JAVA_PACKAGE, "MlflowAutologEventPublisher")
    return getattr(jvm, qualified_classname)


def _get_repl_id():
    """
    Get a unique REPL ID for a PythonSubscriber instance. This is used to distinguish between
    REPLs in multitenant, REPL-aware environments where multiple Python processes may share the
    same Spark JVM (e.g. in Databricks). In such environments, we pull the REPL ID from Spark
    local properties, and expect that the PythonSubscriber for the current Python process only
    receives events for datasource reads triggered by the current process.
    """
    repl_id = databricks_utils.get_repl_id()
    if repl_id:
        return repl_id
    main_file = sys.argv[0] if len(sys.argv) > 0 else "<console>"
    return "PythonSubscriber[{filename}][{id}]".format(
        filename=main_file, id=uuid.uuid4().hex
    )


class PythonSubscriber(object):
    """
    Subscriber, intended to be instantiated once per Python process, that subscribes to Spark for
    information from JVM about the spark data sources read during the lifetime of this subscriber.
    This class implements Java interface (org.mlflow.spark.autologging.MlflowAutologEventSubscriber,
    defined in the mlflow-spark package) that's called-into by autologging logic in the JVM in order
    to propagate Spark datasource read events to Python.

    This class leverages the Py4j callback mechanism to receive callbacks from the JVM, see
    https://www.py4j.org/advanced_topics.html#implementing-java-interfaces-from-python-callback for
    more information.
    """

    def __init__(self):
        self._repl_id = _get_repl_id()
        self._data_sources = []
        self._registered = False
        # Internal message for client, if needed. Can be externalized in future.
        self._message = None

    def toString(self):
        # For debugging
        return "PythonSubscriber<replId=%s>" % self.replId()

    def ping(self):
        return None

    def notify(self, path, version, data_format):
        # Record data sources only if this subscriber is registered.
        if not self.is_active():
            _logger.warning(f"Data source listener not active. Cause: ${self._message}")
        else:
            self._data_sources.append((path, version, data_format))

    def register(self):
        event_publisher = _get_jvm_event_publisher()
        event_publisher.register(self)
        self._registered = True

    def replId(self):
        return self._repl_id

    def get_data_sources(self):
        source_type_to_data_source = {}
        for (path, _, fmt) in self._data_sources:
            if fmt not in source_type_to_data_source:
                source_type_to_data_source[fmt] = []
            source_type_to_data_source[fmt].append(path)
        return source_type_to_data_source

    class Java:
        implements = ["{}.MlflowAutologEventSubscriber".format(_JAVA_PACKAGE)]

    def disable(self):
        self._registered = False

    def set_error(self, msg):
        self._message = msg

    def is_active(self):
        return self._registered


def spark_activity_subscriber(spark_context):
    spark_table_info_listener = PythonSubscriber()
    if _get_spark_major_version(spark_context) < 3:
        msg = "Data source listener disabled. Use Spark version >= 3"
        _logger.warning(msg)
        spark_table_info_listener.set_error(msg)
        return spark_table_info_listener
    gw = spark_context._gateway
    params = gw.callback_server_parameters
    callback_server_params = CallbackServerParameters(
        address=params.address,
        port=params.port,
        daemonize=True,
        daemonize_connections=True,
        eager_load=params.eager_load,
        ssl_context=params.ssl_context,
        accept_timeout=params.accept_timeout,
        read_timeout=params.read_timeout,
        auth_token=params.auth_token,
    )
    callback_server_started = gw.start_callback_server(callback_server_params)

    try:
        event_publisher = _get_jvm_event_publisher()
        event_publisher.init(1)
        spark_table_info_listener.register()
    except Exception as e:
        if callback_server_started:
            try:
                gw.shutdown_callback_server()
            except Exception as e:  # pylint: disable=broad-except
                _logger.warning(
                    "Failed to shut down Spark callback server for autologging: %s",
                    str(e),
                )
        spark_table_info_listener.set_error("Failed to start callback server.")
        spark_table_info_listener.disable()
    if not spark_table_info_listener.is_active() or not _get_active_spark_session():
        _logger.warning(
            "Exception while attempting to initialize JVM-side state for logging "
            "Spark datasource. Either use a cluster with ML Runtime or install "
            "the mlflow-spark JAR in your cluster as described in "
            "http://mlflow.org/docs/latest/tracking.html#automatic-logging-from-spark-experimental."
        )
    return spark_table_info_listener


class SparkSourceListener(object):
    def __init__(self):
        pass

    def __enter__(self):
        active_session = _get_active_spark_session()
        if active_session is not None:
            # We know SparkContext exists here already, so get it
            sc = SparkContext.getOrCreate()
            return spark_activity_subscriber(sc)

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass
